# Copyright (C) 2018 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
S3Storage
==================

A storage provider that stores data in an Amazon S3 bucket.
"""

import io
import logging
from collections import defaultdict
from contextlib import ExitStack
from typing import IO, TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

import boto3
from botocore.exceptions import ClientError

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import Digest
from buildgrid._protos.google.rpc import code_pb2
from buildgrid._protos.google.rpc.status_pb2 import Status
from buildgrid.server.decorators import timed
from buildgrid.server.exceptions import StorageFullError
from buildgrid.server.logging import buildgrid_logger
from buildgrid.server.metrics_names import METRIC
from buildgrid.server.metrics_utils import publish_counter_metric
from buildgrid.server.s3 import s3utils
from buildgrid.server.settings import HASH, S3_MAX_UPLOAD_SIZE, S3_TIMEOUT_READ

from .storage_abc import StorageABC, create_write_session

if TYPE_CHECKING:
    from mypy_boto3_s3 import Client as S3Client

LOGGER = buildgrid_logger(__name__)


class S3Storage(StorageABC):
    TYPE = "S3"

    def __init__(
        self,
        bucket: str,
        page_size: int = 1000,
        s3_read_timeout_seconds_per_kilobyte: Optional[float] = None,
        s3_write_timeout_seconds_per_kilobyte: Optional[float] = None,
        s3_read_timeout_min_seconds: float = S3_TIMEOUT_READ,
        s3_write_timeout_min_seconds: float = S3_TIMEOUT_READ,
        s3_versioned_deletes: bool = False,
        s3_hash_prefix_size: Optional[int] = None,
        s3_path_prefix_string: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        self._bucket_template = bucket
        self._page_size = page_size
        self._s3_read_timeout_seconds_per_kilobyte = s3_read_timeout_seconds_per_kilobyte
        self._s3_read_timeout_min_seconds = s3_read_timeout_min_seconds
        self._s3_write_timeout_seconds_per_kilobyte = s3_write_timeout_seconds_per_kilobyte
        self._s3_write_timeout_min_seconds = s3_write_timeout_min_seconds
        self._s3_versioned_deletes = s3_versioned_deletes
        self._s3_hash_prefix_size = s3_hash_prefix_size
        self._s3_path_prefix_string: Optional[str] = None
        if s3_path_prefix_string:
            self._s3_path_prefix_string = s3_path_prefix_string.strip("/")

        # Boto logs can be very verbose, restrict to WARNING
        for boto_logger_name in ["boto3", "botocore", "s3transfer", "urllib3"]:
            boto_logger = logging.getLogger(boto_logger_name)
            boto_logger.setLevel(max(boto_logger.level, logging.WARNING))

        self._s3: "S3Client" = boto3.client("s3", **kwargs)

    def _construct_key_with_prefix(self, digest: Digest) -> str:
        if not self._s3_hash_prefix_size and not self._s3_path_prefix_string:
            return self._construct_key(digest)
        else:
            try:
                prefix = ""
                if self._s3_path_prefix_string:
                    prefix += self._s3_path_prefix_string + "/"
                if self._s3_hash_prefix_size:
                    prefix += digest.hash[0 : self._s3_hash_prefix_size] + "/"
                    remaining = digest.hash[self._s3_hash_prefix_size :]
                else:
                    remaining = digest.hash
                return f"{prefix}{remaining}_{digest.size_bytes}"
            except IndexError:
                LOGGER.error(
                    (
                        "Could not calculate bucket name for digest. This "
                        "is either a misconfiguration in the BuildGrid S3 bucket "
                        "configuration, or a badly formed request."
                    ),
                    tags=dict(digest=digest),
                )
                raise

    def _get_bucket_name(self, digest: str) -> str:
        try:
            return self._bucket_template.format(digest=digest)
        except IndexError:
            LOGGER.error(
                (
                    "Could not calculate bucket name for digest. This "
                    "is either a misconfiguration in the BuildGrid S3 bucket "
                    "configuration, or a badly formed request."
                ),
                tags=dict(digest=digest),
            )
            raise

    def _construct_key(self, digest: Digest) -> str:
        return digest.hash + "_" + str(digest.size_bytes)

    def _get_s3object(self, digest: Digest) -> s3utils.S3Object:
        object = s3utils.S3Object(self._get_bucket_name(digest.hash), self._construct_key_with_prefix(digest))
        object.filesize = digest.size_bytes
        return object

    def _remove_key_prefixes(self, key: str) -> str:
        # Only interested in last two elements if hash_prefix used
        split_key = key.split("/")
        if self._s3_hash_prefix_size:
            return "".join(split_key[-2:])
        # Only need last element if only a prefix string was used
        if self._s3_path_prefix_string:
            return split_key[-1]
        return key

    def _deconstruct_key(self, key: str) -> Tuple[str, int]:
        # Remove any prefix, returning key to hash_size_bytes format
        key = self._remove_key_prefixes(key)
        parts = key.split("_")
        size_bytes = int(parts[-1])
        # This isn't as simple as just "the first part of the split" because
        # the hash part of the key itself might contain an underscore.
        digest_hash = "_".join(parts[0:-1])
        return digest_hash, size_bytes

    def _multi_delete_blobs(self, bucket_name: str, digests: List[Dict[str, str]]) -> List[str]:
        # TODO fix this:
        #    expression has type "List[Dict[str, str]]",
        #    TypedDict item "Objects" has type "Sequence[ObjectIdentifierTypeDef]"
        response = self._s3.delete_objects(
            Bucket=bucket_name, Delete={"Objects": digests}  # type: ignore[typeddict-item]
        )
        return_failed = []
        failed_deletions = response.get("Errors", [])
        for failed_key in failed_deletions:
            digest_hash, size_bytes = self._deconstruct_key(failed_key["Key"])
            return_failed.append(f"{digest_hash}/{size_bytes}")
        return return_failed

    @timed(METRIC.STORAGE.STAT_DURATION, type=TYPE)
    def has_blob(self, digest: Digest) -> bool:
        LOGGER.debug("Checking for blob.", tags=dict(digest=digest))
        try:
            s3utils.head_object(self._s3, self._get_s3object(digest))
        except ClientError as e:
            if e.response["Error"]["Code"] not in ["404", "NoSuchKey"]:
                raise
            return False
        return True

    @timed(METRIC.STORAGE.READ_DURATION, type=TYPE)
    def get_blob(self, digest: Digest) -> Optional[IO[bytes]]:
        LOGGER.debug("Getting blob.", tags=dict(digest=digest))
        try:
            s3object = self._get_s3object(digest)
            s3object.fileobj = create_write_session(digest)
            s3utils.get_object(
                self._s3,
                s3object,
                timeout_seconds_per_kilobyte=self._s3_read_timeout_seconds_per_kilobyte,
                timeout_min_seconds=self._s3_read_timeout_min_seconds,
            )
            s3object.fileobj.seek(0)
            return s3object.fileobj
        except ClientError as e:
            if e.response["Error"]["Code"] not in ["404", "NoSuchKey"]:
                raise
            return None

    def _get_version_id(self, bucket: str, key: str) -> Optional[str]:
        try:
            return self._s3.head_object(Bucket=bucket, Key=key).get("VersionId")
        except ClientError as e:
            if e.response["Error"]["Code"] not in ["404", "NoSuchKey"]:
                raise
            return None

    @timed(METRIC.STORAGE.DELETE_DURATION, type=TYPE)
    def delete_blob(self, digest: Digest) -> None:
        LOGGER.debug("Deleting blob.", tags=dict(digest=digest))
        bucket, key = self._get_bucket_name(digest.hash), self._construct_key_with_prefix(digest)
        try:
            if self._s3_versioned_deletes and (version_id := self._get_version_id(bucket, key)):
                self._s3.delete_object(Bucket=bucket, Key=key, VersionId=version_id)
            else:
                self._s3.delete_object(Bucket=bucket, Key=key)
        except ClientError as e:
            if e.response["Error"]["Code"] not in ["404", "NoSuchKey"]:
                raise

    def _get_version_ids(self, digests: List[Digest]) -> Dict[str, Optional[str]]:
        s3objects = [self._get_s3object(digest) for digest in digests]
        s3utils.head_objects(self._s3, s3objects)
        return {
            digest.hash: s3object.response_headers.get("x-amz-version-id")
            for digest, s3object in zip(digests, s3objects)
            if s3object.error is None
        }

    @timed(METRIC.STORAGE.BULK_DELETE_DURATION, type=TYPE)
    def bulk_delete(self, digests: List[Digest]) -> List[str]:
        # Avoid expensive string creation
        if LOGGER.is_enabled_for(logging.DEBUG):
            LOGGER.debug(f"Deleting {len(digests)} digests from S3 storage: [{digests}]")

        if self._s3_versioned_deletes:
            version_ids = self._get_version_ids(digests)
        else:
            version_ids = {}

        bucketed_requests: Dict[str, List[Dict[str, str]]] = defaultdict(list)
        for digest in digests:
            bucket = self._get_bucket_name(digest.hash)
            key = self._construct_key_with_prefix(digest)
            bucketed_requests[bucket].append({"Key": key})
            if version_id := version_ids.get(digest.hash):
                bucketed_requests[bucket].append({"Key": key, "VersionId": version_id})

        failed_deletions = []
        for bucket, requests in bucketed_requests.items():
            for i in range(0, len(requests), self._page_size):
                try:
                    failed_deletions += self._multi_delete_blobs(bucket, requests[i : i + self._page_size])
                except ClientError as error:
                    current_failed_deletions = [
                        self._deconstruct_key(key_versions["Key"])
                        for key_versions in requests[i : i + self._page_size]
                    ]
                    failed_deletions += [
                        f"{digest_hash}/{digest_size_bytes}"
                        for digest_hash, digest_size_bytes in current_failed_deletions
                    ]
                    LOGGER.exception(error)
                    LOGGER.exception("Error encountered when trying to delete blobs from the S3 storage.")

        publish_counter_metric(METRIC.STORAGE.DELETE_ERRORS_COUNT, len(failed_deletions), type=self.TYPE)
        return failed_deletions

    @timed(METRIC.STORAGE.WRITE_DURATION, type=TYPE)
    def commit_write(self, digest: Digest, write_session: IO[bytes]) -> None:
        LOGGER.debug("Writing blob.", tags=dict(digest=digest))
        write_session.seek(0)
        try:
            s3object = self._get_s3object(digest)
            s3object.fileobj = write_session
            s3object.filesize = digest.size_bytes
            if digest.size_bytes <= S3_MAX_UPLOAD_SIZE:
                s3utils.put_object(
                    self._s3,
                    s3object,
                    timeout_seconds_per_kilobyte=self._s3_write_timeout_seconds_per_kilobyte,
                    timeout_min_seconds=self._s3_write_timeout_min_seconds,
                )
            else:
                s3utils.multipart_upload(self._s3, s3object)
        except ClientError as error:
            if error.response["Error"]["Code"] == "QuotaExceededException":
                raise StorageFullError("S3 Quota Exceeded.") from error
            raise error

    @timed(METRIC.STORAGE.BULK_STAT_DURATION, type=TYPE)
    def missing_blobs(self, digests: List[Digest]) -> List[Digest]:
        result = []
        s3objects = []
        for digest in digests:
            s3object = self._get_s3object(digest)
            s3objects.append(s3object)
        s3utils.head_objects(self._s3, s3objects)
        for digest, s3object in zip(digests, s3objects):
            if s3object.error is not None:
                result.append(digest)
        return result

    @timed(METRIC.STORAGE.BULK_WRITE_DURATION, type=TYPE)
    def bulk_update_blobs(self, blobs: List[Tuple[Digest, bytes]]) -> List[Status]:
        s3object_status_list: List[Union[Tuple[None, Status], Tuple[s3utils.S3Object, None]]] = []
        s3objects = []
        with ExitStack() as stack:
            for digest, data in blobs:
                if len(data) != digest.size_bytes or HASH(data).hexdigest() != digest.hash:
                    status = Status(
                        code=code_pb2.INVALID_ARGUMENT,
                        message="Data doesn't match hash",
                    )
                    s3object_status_list.append((None, status))
                else:
                    write_session = stack.enter_context(create_write_session(digest))
                    write_session.write(data)
                    write_session.seek(0)
                    s3object = self._get_s3object(digest)
                    s3object.fileobj = write_session
                    s3object.filesize = digest.size_bytes
                    s3objects.append(s3object)
                    s3object_status_list.append((s3object, None))

            s3utils.put_objects(
                self._s3,
                s3objects,
                timeout_seconds_per_kilobyte=self._s3_write_timeout_seconds_per_kilobyte,
                timeout_min_seconds=self._s3_write_timeout_min_seconds,
            )

            result = []
            for res_s3object, res_status in s3object_status_list:
                if res_status:
                    # Failed check before S3 object creation
                    result.append(res_status)
                elif res_s3object:
                    if res_s3object.error is None:
                        # PUT was successful
                        result.append(Status(code=code_pb2.OK))
                    else:
                        result.append(Status(code=code_pb2.UNKNOWN, message=str(res_s3object.error)))

        return result

    @timed(METRIC.STORAGE.BULK_READ_DURATION, type=TYPE)
    def bulk_read_blobs(self, digests: List[Digest]) -> Dict[str, bytes]:
        s3objects: List[s3utils.S3Object] = []
        for digest in digests:
            s3object = self._get_s3object(digest)
            s3object.fileobj = io.BytesIO()
            s3objects.append(s3object)

        s3utils.get_objects(
            self._s3,
            s3objects,
            timeout_seconds_per_kilobyte=self._s3_read_timeout_seconds_per_kilobyte,
            timeout_min_seconds=self._s3_read_timeout_min_seconds,
        )

        blobmap: Dict[str, bytes] = {}
        for digest, s3object in zip(digests, s3objects):
            if not s3object.error:
                if s3object.fileobj:
                    s3object.fileobj.seek(0)
                    blobmap[digest.hash] = s3object.fileobj.read()
            elif s3object.status_code != 404:
                raise s3object.error
        return blobmap



# Copyright (C) 2018 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Execution Controller
====================

An instance of the Execution controller.

All this stuff you need to make the execution service work.

Contains scheduler, execution instance, an interface to the bots
and an operations instance.
"""
from typing import Iterable, Optional

from buildgrid.server.bots.instance import BotsInterface
from buildgrid.server.enums import ServiceName
from buildgrid.server.execution.instance import ExecutionInstance
from buildgrid.server.operations.instance import OperationsInstance
from buildgrid.server.scheduler import Scheduler
from buildgrid.server.settings import DEFAULT_MAX_LIST_OPERATION_PAGE_SIZE


class ExecutionController:
    def __init__(
        self,
        scheduler: Scheduler,
        *,
        operation_stream_keepalive_timeout: Optional[int] = None,
        services: Iterable[str] = ServiceName.default_services(),
        max_list_operations_page_size: int = DEFAULT_MAX_LIST_OPERATION_PAGE_SIZE,
    ) -> None:
        self.execution_instance: Optional[ExecutionInstance] = None
        if ServiceName.EXECUTION.value in services:
            self.execution_instance = ExecutionInstance(scheduler, operation_stream_keepalive_timeout)

        self.operations_instance: Optional[OperationsInstance] = None
        if ServiceName.OPERATIONS.value in services:
            self.operations_instance = OperationsInstance(scheduler, max_list_operations_page_size)

        self.bots_interface: Optional[BotsInterface] = None
        if ServiceName.BOTS.value in services:
            self.bots_interface = BotsInterface(scheduler)



# Copyright (C) 2019 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
IndexABC
==================

The abstract base class for storage indices. An index is a special type of
Storage that facilitates storing blob metadata. It must wrap another Storage.

Derived classes must implement all methods of both this interface and the
StorageABC interface.
"""

import abc
from datetime import datetime
from typing import Iterator, Optional

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import Digest

from ..storage_abc import StorageABC


class IndexABC(StorageABC):
    @abc.abstractmethod
    def __init__(self, *, fallback_on_get: bool = False) -> None:
        # If fallback is enabled, the index is required to fetch blobs from
        # storage on each get_blob and bulk_read_blobs request and update
        # itself accordingly.
        self._fallback_on_get = fallback_on_get

    @abc.abstractmethod
    def least_recent_digests(self) -> Iterator[Digest]:
        """Generator to iterate through the digests in LRU order"""

    @abc.abstractmethod
    def get_total_size(self) -> int:
        """
        Return the sum of the size of all blobs within the index.
        """

    @abc.abstractmethod
    def delete_n_bytes(
        self,
        n_bytes: int,
        dry_run: bool = False,
        protect_blobs_after: Optional[datetime] = None,
    ) -> int:
        """Delete around n bytes of data from the index.

        The ordering in which digests are deleted is up to the specific implementations
        and may provide different semantics such as LRU or random deletion. ALL implementations
        must respect the protect_blobs_after parameter to limit the age of deleted blobs

        Implementations should generate delete around n_bytes on each call, but may delete more or less
        depending on the state of the storage, index, and value of protect_blobs_after.

        Args:
            n_bytes (int): The number of bytes to be deleted
            dry_run (bool): Don't actually delete any data, just return the number of bytes which would be deleted
            protect_blobs_after: Don't delete any digests which have been accessed after this time

        Returns:
            Number of bytes deleted
        """



# Copyright (C) 2021 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import re
from threading import RLock
from typing import Dict, Iterator, List, Optional

from buildgrid._protos.buildgrid.v2.query_build_events_pb2 import QueryEventStreamsResponse
from buildgrid._protos.google.devtools.build.v1.build_events_pb2 import StreamId
from buildgrid._protos.google.devtools.build.v1.publish_build_event_pb2 import OrderedBuildEvent
from buildgrid.server.logging import buildgrid_logger
from buildgrid.server.servicer import Instance

LOGGER = buildgrid_logger(__name__)


class DuplicateStreamError(Exception):
    """Error when encountering a name collision between event streams."""

    pass


class BuildEventStream:
    """Internal representation of a stream of OrderedBuildEvents.

    This class provides in-memory storage of the events in a given Build Event
    Stream. Many of these Build Event Streams may relate to the same build, or
    even to the same invocation of a tool.

    Args:
        stream_id (StreamId): A gRPC message defining the ID of this stream.

    """

    def __init__(self, stream_id: StreamId):
        self._stream_id = stream_id
        self._events: List[OrderedBuildEvent] = []
        LOGGER.debug("Created BuildEventStream.", tags=dict(stream_id=self._stream_id))

    def __len__(self) -> int:
        return len(self._events)

    def publish_event(self, event: OrderedBuildEvent) -> None:
        """Publish an ``OrderedBuildEvent`` into the stream.

        Args:
            event (OrderedBuildEvent): The event to publish to the stream.
                This is an ``OrderedBuildEvent`` message from the
                ``publish_build_event`` proto.
        """
        self._events.append(event)
        LOGGER.debug("Stored BuildEvent in stream.", tags=dict(stream_id=self._stream_id))

    def query_events(self, query: Optional[str] = None) -> Iterator[OrderedBuildEvent]:
        """Query the contents of this stream.

        Filter the contents of the event stream by some query, returning an
        iterator of matching OrderedBuildEvents.

        .. note::
            The filtering functionality of this method is currently not
            implemented, and the iterator returned contains all the events
            in the stream no matter what query is used.

        Args:
            query (str): The filter string to use when querying events.
        """
        # TODO(SotK): Implement some basic querying here
        if query is not None:
            raise NotImplementedError("Specifying a build events query is not supported yet.")
        yield from self._events

    def to_grpc_message(self) -> QueryEventStreamsResponse.BuildEventStream:
        """Convert this object to a ``BuildEventStream`` gRPC message.

        This method converts this internal event stream representation into
        a ``QueryEventStreamsResponse.BuildEventStream`` gRPC message, as
        defined in the ``query_build_events`` proto.

        """
        return QueryEventStreamsResponse.BuildEventStream(stream_id=self._stream_id, events=self._events)


class BuildEventStreamStorage(Instance):
    """In-memory storage of Build Event Streams.

    This class stores a collection of Build Event Streams, and handles both
    creation of new streams and querying the streams which already exist
    based on their stream ID. Streams are stored in-memory and are lost on
    service restart, so shouldn't be relied on as a source of persistent
    data when using this storage class.

    This class is similar to the ``.*Instance`` classes used by other
    BuildGrid services, in that it is instantiated by the config parser
    and used by a ``.*Service`` class instantiated by the server class.
    Unlike the other instance classes however, this class doesn't have an
    ``instance_name`` attribute due to the Build Events protos not having
    the concept of multiple instances.

    """

    # Hack in a service name here to fit the pattern
    SERVICE_NAME = "BuildEvents"

    def __init__(self) -> None:
        self._streams: Dict[str, BuildEventStream] = {}
        self._streams_lock = RLock()

    def new_stream(self, stream_id: StreamId) -> BuildEventStream:
        """Create and return a new ``BuildEventStream`` with the given ID.

        This method creates a new :class:`BuildEventStream` with the given
        ``StreamId``, and returns it. If a stream with that ID already exists
        in this ``BuildEventStreamStorage``, then a :class:`DuplicateStreamError`
        is raised.

        Args:
            stream_id (StreamId): The gRPC StreamId message containing the
                ID of the stream to create.

        """
        key = self._get_stream_key(stream_id)
        with self._streams_lock:
            if key in self._streams:
                raise DuplicateStreamError(f"Stream with key {key} already exists.")

            stream = BuildEventStream(stream_id)
            self._streams[key] = stream
            return stream

    def get_stream(self, stream_id: StreamId) -> BuildEventStream:
        """Return a ``BuildEventStream`` with the given stream ID.

        This method takes a stream ID, converts it to a stream key, and
        returns the stream with that key if one exists.

        This method will create a new :class:`BuildEventStream` if one with
        the given ID doesn't exist.

        Args:
            stream_id (StreamId): The StreamID message from an event to
                get the ``BuildEventStream`` for.

        Returns:
            A ``BuildEventStream`` with the given ID, or None.
        """
        key = self._get_stream_key(stream_id)
        with self._streams_lock:
            stream = self._streams.get(key)
            if stream is None:
                stream = self.new_stream(stream_id)
            return stream

    def get_matching_streams(self, stream_key_regex: str) -> List[BuildEventStream]:
        """Return a list of event streams which match the given key pattern.

        This method takes a regular expression as a string, and returns
        a list of :class:`BuildEventStream` objects whose stream key (based
        on the events' StreamId) matches the regex.

        Args:
            stream_key_regex (str): A regular expression to use to find
                matching streams.

        Returns:
            List of :class:`BuildEventStream` objects which have a key
            matching the given regex.

        """
        regex = re.compile(stream_key_regex)
        return [stream for key, stream in self._streams.items() if regex.search(key) is not None]

    def _get_stream_key(self, stream_id: StreamId) -> str:
        return f"{stream_id.build_id}.{stream_id.component}.{stream_id.invocation_id}"



# Copyright (C) 2020 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# flake8: noqa
# fmt: off
class METRIC:
    class RPC:
        DURATION                            = "rpc.duration.ms"
        INPUT_BYTES                         = "rpc.input_bytes.count"
        OUTPUT_BYTES                        = "rpc.output_bytes.count"
        AUTH_DURATION                       = "rpc.auth.duration.ms"

    class ACTION_CACHE:
        INVALID_CACHE_COUNT                 = "action_cache.invalid_cache.count"
        MIRRORED_MATCH_COUNT                = "action_cache.mirrored_matches.count"
        MIRRORED_MISMATCH_COUNT             = "action_cache.mirrored_mismatches.count"
        RESULT_AGE                          = "action_cache.result_age.ms"

    class CAS:
        BLOBS_COUNT                         = "cas.blobs.count"
        BLOBS_MISSING_COUNT                 = "cas.blobs_missing.count"
        BLOBS_MISSING_PERCENT               = "cas.blobs_missing.percent"
        BLOB_BYTES                          = "cas.blob_bytes.count"
        TREE_CACHE_HIT_COUNT                = "cas.tree_cache_hit.count"
        TREE_CACHE_MISS_COUNT               = "cas.tree_cache_miss.count"

    class STORAGE:
        STAT_DURATION                       = "storage.stat.duration.ms"
        BULK_STAT_DURATION                  = "storage.bulk_stat.duration.ms"

        READ_DURATION                       = "storage.read.duration.ms"
        BULK_READ_DURATION                  = "storage.bulk_read.duration.ms"

        DELETE_DURATION                     = "storage.delete_blob.duration.ms"
        BULK_DELETE_DURATION                = "storage.bulk_delete.duration.ms"
        DELETE_ERRORS_COUNT                 = "storage.delete_errors.count"

        WRITE_DURATION                      = "storage.write.duration.ms"
        BULK_WRITE_DURATION                 = "storage.bulk_write.duration.ms"

        GET_TREE_DURATION                   = "storage.get_tree.duration.ms"

        class WITH_CACHE:
            CACHE_HIT_COUNT                 = "storage.with_cache.cache_hit.count"
            CACHE_MISS_COUNT                = "storage.with_cache.cache_miss.count"
            CACHE_HIT_PERCENT               = "storage.with_cache.cache_hit.percent"

        class SQL_INDEX:
            UPDATE_TIMESTAMP_DURATION       = "storage.sql_index.update_timestamp.duration.ms"
            SAVE_DIGESTS_DURATION           = "storage.sql_index.save_digest.duration.ms"
            SIZE_CALCULATION_DURATION       = "storage.sql_index.size_calculation.duration.ms"
            DELETE_N_BYTES_DURATION         = "storage.sql_index.delete_n_bytes.duration.ms"
            BULK_DELETE_INDEX_DURATION      = "storage.sql_index.bulk_delete_index.duration.ms"
            MARK_DELETED_DURATION           = "storage.sql_index.mark_deleted.duration.ms"
            PREMARKED_DELETED_COUNT         = "storage.sql_index.premarked_deleted.count"

        class REPLICATED:
            REQUIRED_REPLICATION_COUNT      = "storage.replicated.required_replication.count"
            REPLICATION_COUNT               = "storage.replicated.replication.count"
            REPLICATION_ERROR_COUNT         = "storage.replicated.replication.errors.count"

    class CLEANUP:
        DURATION                            = "cleanup.duration.ms"
        BATCH_DURATION                      = "cleanup.batch.duration.ms"
        BLOBS_DELETED_PER_SECOND            = "cleanup.blobs_deleted.per_second"
        BYTES_DELETED_PER_SECOND            = "cleanup.bytes_deleted.per_second"
        BYTES_DELETED_COUNT                 = "cleanup.bytes_deleted.count"
        TOTAL_BYTES_COUNT                   = "cleanup.total_bytes.count"
        LOW_WATERMARK_BYTES_COUNT           = "cleanup.low_watermark_bytes.count"
        HIGH_WATERMARK_BYTES_COUNT          = "cleanup.high_watermark_bytes.count"

    class SCHEDULER:
        JOB_COUNT                           = "scheduler.jobs.count"
        LEASE_COUNT                         = "scheduler.leases.count"
        BOTS_COUNT                          = "scheduler.bots.count"

        ASSIGNMENT_DURATION                 = "scheduler.assignment.duration.ms"
        SYNCHRONIZE_DURATION                = "scheduler.synchronize.duration.ms"

        PRUNE_DURATION                      = "scheduler.prune.duration.ms"
        PRUNE_COUNT                         = "scheduler.prune.count"

        QUEUE_TIMEOUT_DURATION              = "scheduler.queue_timeout.duration.ms"
        QUEUE_TIMEOUT_COUNT                 = "scheduler.queue_timeout.count"

        EXECUTION_TIMEOUT_DURATION          = "scheduler.execution_timeout.duration.ms"
        EXECUTION_TIMEOUT_COUNT             = "scheduler.execution_timeout.count"

    class CONNECTIONS:
        CLIENT_COUNT                        = "connections.clients.count"
        WORKER_COUNT                        = "connections.workers.count"

    class JOB:
        DURATION                            = "job.duration.ms"



# Copyright (C) 2024 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import inspect
from typing import Any, Callable, Iterator, TypeVar, cast

import grpc

from buildgrid.server.auth.manager import authorize_request
from buildgrid.server.context import current_instance
from buildgrid.server.exceptions import InvalidArgumentError
from buildgrid.server.metrics_names import METRIC
from buildgrid.server.metrics_utils import timer

Func = TypeVar("Func", bound=Callable)  # type: ignore[type-arg]


def authorize(f: Func) -> Func:
    @functools.wraps(f)
    def server_stream_wrapper(self: Any, message: Any, context: grpc.ServicerContext) -> Iterator[Any]:
        try:
            with timer(METRIC.RPC.AUTH_DURATION):
                authorize_request(context, current_instance(), f.__name__)
            yield from f(self, message, context)
        except InvalidArgumentError as e:
            context.abort(grpc.StatusCode.INVALID_ARGUMENT, str(e))

    @functools.wraps(f)
    def server_unary_wrapper(self: Any, message: Any, context: grpc.ServicerContext) -> Any:
        try:
            with timer(METRIC.RPC.AUTH_DURATION):
                authorize_request(context, current_instance(), f.__name__)
            return f(self, message, context)
        except InvalidArgumentError as e:
            context.abort(grpc.StatusCode.INVALID_ARGUMENT, str(e))

    if inspect.isgeneratorfunction(f):
        return cast(Func, server_stream_wrapper)
    return cast(Func, server_unary_wrapper)



# Copyright (C) 2019 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from logging import getLogger
from logging.config import fileConfig

from alembic import context
from sqlalchemy import engine_from_config, pool
from sqlalchemy.engine.base import Connection

from buildgrid.server.sql.models import Base

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

# Interpret the config file for Python logging.
# This line sets up loggers basically.
try:
    fileConfig(config.config_file_name)  # type: ignore[arg-type]  # covariance in union issue. Not a real error.
except TypeError:
    pass

LOGGER = getLogger(__name__)

target_metadata = Base.metadata

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def run_migrations_offline() -> None:  # pragma: no cover
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    url = config.get_main_option("sqlalchemy.url")
    context.configure(url=url, target_metadata=target_metadata, literal_binds=True)

    with context.begin_transaction():
        context.run_migrations()


def _run_migrations_online(connection: Connection) -> None:  # pragma: no cover
    context.configure(connection=connection, target_metadata=target_metadata)

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online() -> None:  # pragma: no cover
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    try:
        connection = config.attributes["connection"]
        _run_migrations_online(connection)

    except KeyError:
        config_section = config.get_section(config.config_ini_section)
        if config_section is not None:
            connectable = engine_from_config(config_section, prefix="sqlalchemy.", poolclass=pool.NullPool)
            with connectable.connect() as connection:
                _run_migrations_online(connection)
        else:
            LOGGER.error(
                "Configuration file must define either `connection` or SQLAlchemy config "
                f"keys prefixed with `sqlalchemy` in the {config.config_ini_section} section."
            )
            sys.exit(1)


if context.is_offline_mode():  # pragma: no cover
    run_migrations_offline()
else:  # pragma: no cover
    run_migrations_online()



# Copyright (C) 2018 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import grpc

from buildgrid._protos.build.bazel.remote.execution.v2 import remote_execution_pb2, remote_execution_pb2_grpc
from buildgrid.server.logging import buildgrid_logger

LOGGER = buildgrid_logger(__name__)


class CapabilitiesInterface:
    """Interface for calls the the Capabilities Service."""

    def __init__(self, channel: grpc.Channel) -> None:
        """Initialises an instance of the capabilities service.

        Args:
            channel (grpc.Channel): A gRPC channel to the CAS endpoint.
        """
        self.__stub = remote_execution_pb2_grpc.CapabilitiesStub(channel)

    def get_capabilities(self, instance_name: str) -> remote_execution_pb2.ServerCapabilities:
        """Returns the capabilities or the server to the user.

        Args:
            instance_name (str): The name of the instance."""

        request = remote_execution_pb2.GetCapabilitiesRequest(instance_name=instance_name)
        try:
            return self.__stub.GetCapabilities(request)

        except grpc.RpcError as e:
            LOGGER.debug(e)
            raise ConnectionError(e.details())



# Copyright (C) 2023 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Add platform properties table

Revision ID: afa46425330d
Revises: 9b595964dc25
Create Date: 2023-10-05 17:08:45.628326

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "afa46425330d"
down_revision = "9b595964dc25"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "platform_properties",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("key", sa.String(), nullable=True),
        sa.Column("value", sa.String(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("key", "value"),
    )
    op.create_table(
        "job_platforms",
        sa.Column("job_name", sa.String(), nullable=False),
        sa.Column("platform_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["job_name"],
            ["jobs.name"],
        ),
        sa.ForeignKeyConstraint(
            ["platform_id"],
            ["platform_properties.id"],
        ),
        sa.PrimaryKeyConstraint("job_name", "platform_id"),
    )


def downgrade() -> None:
    op.drop_table("job_platforms")
    op.drop_table("platform_properties")



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2018 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.descriptor_pb2
import google.protobuf.internal.containers
import google.protobuf.internal.extension_dict

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

METHOD_SIGNATURE_FIELD_NUMBER: builtins.int
DEFAULT_HOST_FIELD_NUMBER: builtins.int
OAUTH_SCOPES_FIELD_NUMBER: builtins.int
method_signature: google.protobuf.internal.extension_dict._ExtensionFieldDescriptor[google.protobuf.descriptor_pb2.MethodOptions, google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]]
"""A definition of a client library method signature.

In client libraries, each proto RPC corresponds to one or more methods
which the end user is able to call, and calls the underlying RPC.
Normally, this method receives a single argument (a struct or instance
corresponding to the RPC request object). Defining this field will
add one or more overloads providing flattened or simpler method signatures
in some languages.

The fields on the method signature are provided as a comma-separated
string.

For example, the proto RPC and annotation:

  rpc CreateSubscription(CreateSubscriptionRequest)
      returns (Subscription) {
    option (google.api.method_signature) = "name,topic";
  }

Would add the following Java overload (in addition to the method accepting
the request object):

  public final Subscription createSubscription(String name, String topic)

The following backwards-compatibility guidelines apply:

  * Adding this annotation to an unannotated method is backwards
    compatible.
  * Adding this annotation to a method which already has existing
    method signature annotations is backwards compatible if and only if
    the new method signature annotation is last in the sequence.
  * Modifying or removing an existing method signature annotation is
    a breaking change.
  * Re-ordering existing method signature annotations is a breaking
    change.
"""
default_host: google.protobuf.internal.extension_dict._ExtensionFieldDescriptor[google.protobuf.descriptor_pb2.ServiceOptions, builtins.str]
"""The hostname for this service.
This should be specified with no prefix or protocol.

Example:

  service Foo {
    option (google.api.default_host) = "foo.googleapi.com";
    ...
  }
"""
oauth_scopes: google.protobuf.internal.extension_dict._ExtensionFieldDescriptor[google.protobuf.descriptor_pb2.ServiceOptions, builtins.str]
"""OAuth scopes needed for the client.

Example:

  service Foo {
    option (google.api.oauth_scopes) = \\
      "https://www.googleapis.com/auth/cloud-platform";
    ...
  }

If there is more than one scope, use a comma-separated string:

Example:

  service Foo {
    option (google.api.oauth_scopes) = \\
      "https://www.googleapis.com/auth/cloud-platform,"
      "https://www.googleapis.com/auth/monitoring";
    ...
  }
"""



# Copyright (C) 2020 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys
from typing import Any, Iterator

import click
from grpc import RpcError

from buildgrid._protos.build.bazel.remote.logstream.v1.remote_logstream_pb2 import CreateLogStreamRequest
from buildgrid._protos.build.bazel.remote.logstream.v1.remote_logstream_pb2_grpc import LogStreamServiceStub
from buildgrid._protos.google.bytestream.bytestream_pb2 import QueryWriteStatusRequest, ReadRequest, WriteRequest
from buildgrid._protos.google.bytestream.bytestream_pb2_grpc import ByteStreamStub
from buildgrid.server.client.channel import setup_channel

from ..cli import pass_context


def instanced_resource_name(instance_name: str, resource: str) -> str:
    """Prepend the resource name with the instance name if it is provided."""
    if instance_name:
        return f"{instance_name}/{resource}"
    return resource


@click.group(name="logstream", short_help="LogStream commands.")
@click.option(
    "--remote",
    type=click.STRING,
    default="http://localhost:50051",
    show_default=True,
    help="Remote execution server's URL (port defaults to 50051 if no specified).",
)
@click.option("--instance-name", type=click.STRING, default="", show_default=True, help="Targeted farm instance name.")
@click.option(
    "--auth-token",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Authorization token for the remote.",
)
@click.option(
    "--client-key",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Private client key for TLS (PEM-encoded).",
)
@click.option(
    "--client-cert",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Public client certificate for TLS (PEM-encoded).",
)
@click.option(
    "--server-cert",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Public server certificate for TLS (PEM-encoded).",
)
@pass_context
def cli(
    context: Any, remote: str, instance_name: str, auth_token: str, client_key: str, client_cert: str, server_cert: str
) -> None:
    context.channel, _ = setup_channel(
        remote, auth_token=auth_token, client_key=client_key, client_cert=client_cert, server_cert=server_cert
    )
    context.instance_name = instance_name


@cli.command("create", short_help="Create a LogStream.")
@click.argument("parent", nargs=1, type=click.STRING, required=True)
@pass_context
def create(context: Any, parent: str) -> None:
    stub = LogStreamServiceStub(context.channel)

    parent = instanced_resource_name(context.instance_name, parent)

    request = CreateLogStreamRequest(parent=parent)
    try:
        logstream = stub.CreateLogStream(request)
    except RpcError as e:
        click.echo(f"Error: {e.details()}", err=True)
        sys.exit(-1)

    click.echo(logstream)


@cli.command("read", short_help="Read from a LogStream.")
@click.argument("resource", nargs=1, type=click.STRING, required=True)
@pass_context
def read(context: Any, resource: str) -> None:
    stub = ByteStreamStub(context.channel)

    resource = instanced_resource_name(context.instance_name, resource)

    request = ReadRequest(resource_name=resource)
    try:
        for response in stub.Read(request):
            click.echo(response.data, nl=False)
    except RpcError as e:
        click.echo(f"Error: {e.details()}", err=True)
        sys.exit(-1)


@cli.command("write", short_help="Write to a LogStream.")
@click.argument("resource", nargs=1, type=click.STRING, required=True)
@click.option("--wait-for-reader", is_flag=True)
@pass_context
def write(context: Any, resource: str, wait_for_reader: bool) -> None:
    stub = ByteStreamStub(context.channel)

    resource = instanced_resource_name(context.instance_name, resource)

    if wait_for_reader:
        try:
            query_write_status = QueryWriteStatusRequest(resource_name=resource)
            stub.QueryWriteStatus(query_write_status)
        except RpcError as e:
            click.echo(f"Error calling QueryWriteStatus: {e.details()}", err=True)
            sys.exit(-1)

    def _write_request_stream(name: str) -> Iterator[WriteRequest]:
        offset = 0
        finished = False
        while not finished:
            line = input("log line (empty line ends the stream): ").strip()
            finished = line == ""

            request = WriteRequest()
            request.resource_name = name
            request.data = line.encode("utf-8")
            request.write_offset = offset
            request.finish_write = finished

            yield request

            offset += 1

    try:
        requests = _write_request_stream(resource)
        stub.Write(requests)
    except RpcError as e:
        click.echo(f"Writing to stream failed: {e.details()}", err=True)
        sys.exit(-1)



# Copyright (C) 2019 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
SQLIndex
==================

A SQL index implementation. This can be pointed to either a remote SQL server
or a local SQLite database.

"""
import io
import itertools
import time
from collections import deque
from datetime import datetime, timedelta
from typing import IO, Any, AnyStr, Deque, Dict, Iterator, List, Optional, Sequence, Tuple, cast

from sqlalchemy import ColumnElement, Table, and_, delete, func, not_, select
from sqlalchemy.exc import DBAPIError
from sqlalchemy.orm import InstrumentedAttribute, Session, load_only
from sqlalchemy.orm.exc import StaleDataError
from sqlalchemy.orm.query import Query
from sqlalchemy.orm.session import Session as SessionType

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import Digest
from buildgrid._protos.google.rpc import code_pb2
from buildgrid._protos.google.rpc.status_pb2 import Status
from buildgrid.server.decorators import timed
from buildgrid.server.exceptions import StorageFullError
from buildgrid.server.logging import buildgrid_logger
from buildgrid.server.metrics_names import METRIC
from buildgrid.server.metrics_utils import publish_counter_metric, publish_gauge_metric, publish_timer_metric
from buildgrid.server.sql.models import IndexEntry
from buildgrid.server.sql.provider import SqlProvider
from buildgrid.server.utils.digests import validate_digest_data

from ..storage_abc import StorageABC
from .index_abc import IndexABC
from .sql_dialect_delegates import PostgreSQLDelegate, SQLiteDelegate

LOGGER = buildgrid_logger(__name__)
DIALECT_DELEGATES = {"postgresql": PostgreSQLDelegate, "sqlite": SQLiteDelegate}

INLINE_BLOB_SIZE_HARD_MAXIMUM = 1000000000


def read_and_rewind(read_head: IO[AnyStr]) -> Optional[AnyStr]:
    """Reads from an IO object and returns the data found there
    after rewinding the object to the beginning.

    Args:
        read_head (IO): readable IO head

    Returns:
        AnyStr: readable content from `read_head`.
    """
    if not read_head:
        return None

    data = read_head.read()
    read_head.seek(0)
    return data


class SQLIndex(IndexABC):
    TYPE = "SQLIndex"

    def __init__(
        self,
        sql_provider: SqlProvider,
        storage: StorageABC,
        *,
        window_size: int = 1000,
        inclause_limit: int = -1,
        max_inline_blob_size: int = 0,
        refresh_accesstime_older_than: int = 0,
        **kwargs: Any,
    ) -> None:
        base_argnames = ["fallback_on_get"]
        base_args = {}
        for arg in base_argnames:
            if arg in kwargs:
                base_args[arg] = kwargs.pop(arg)
        super().__init__(**base_args)

        self._sql = sql_provider
        self._storage = storage

        if max_inline_blob_size > INLINE_BLOB_SIZE_HARD_MAXIMUM:
            raise ValueError(
                f"Max inline blob size is [{max_inline_blob_size}], "
                f"but must be less than [{INLINE_BLOB_SIZE_HARD_MAXIMUM}]."
            )
        if max_inline_blob_size >= 0:
            self._max_inline_blob_size = max_inline_blob_size
        else:
            raise ValueError(f"Max inline blob size is [{max_inline_blob_size}], but must be nonnegative.")

        if refresh_accesstime_older_than >= 0:
            # Measured in seconds. Helps reduce the frequency of timestamp updates during heavy read.
            self.refresh_accesstime_older_than = refresh_accesstime_older_than
        else:
            raise ValueError(
                f"'refresh_accesstime_older_than' must be nonnegative. It is {refresh_accesstime_older_than}"
            )

        # Max # of rows to fetch while iterating over all blobs
        # (e.g. in least_recent_digests)
        self._all_blobs_window_size = window_size

        # This variable stores the list of whereclauses (SQLAlchemy BooleanClauseList objects)
        # generated from the _column_windows() using time-expensive SQL query.
        # These whereclauses are used to construct the final SQL query
        # during cleanup in order to fetch blobs by time windows.
        #
        # Inside the _column_windows() a list of timestamp boarders are obtained:
        #   intervals = [t1, t2, t3, ...]
        # Then the generated whereclauses might represent semantically as, for example,:
        #   self._queue_of_whereclauses = [
        #     "WHERE t1 <= IndexEntry.accessed_timestamp < t2",
        #     "WHERE t2 <= IndexEntry.accessed_timestamp < t3",
        #     "WHERE t3 <= IndexEntry.accessed_timestamp < t4",
        #     ... and so on]
        # Note the number of entries in each window is determined by
        # the instance variable "_all_blobs_window_size".
        self._queue_of_whereclauses: Deque[ColumnElement[bool]] = deque()

        # Whether entries with deleted=True should be considered by mark_n_bytes.
        # This is useful to catch any half-finished deletes where a cleanup process
        # may have exited half-way through deletion. Once all premarked blobs have been
        # deleted this becomes False and is only reset after a full scan of the database
        self._delete_premarked_blobs: bool = True

        # Only pass known kwargs to db session
        available_options = {"pool_size", "max_overflow", "pool_timeout", "pool_pre_ping", "pool_recycle"}
        kwargs_keys = kwargs.keys()
        if kwargs_keys > available_options:
            unknown_args = kwargs_keys - available_options
            raise TypeError(f"Unknown keyword arguments: [{unknown_args}]")

        # Dialect-specific initialization
        self._dialect_delegate = DIALECT_DELEGATES.get(self._sql.dialect)

        if inclause_limit > 0:
            if inclause_limit > window_size:
                LOGGER.warning(
                    "Configured inclause limit is greater than window size.",
                    tags=dict(inclause_limit=inclause_limit, window_size=window_size),
                )
            self._inclause_limit = inclause_limit
        else:
            # If the inlimit isn't explicitly set, we use a default that
            # respects both the window size and the db implementation's
            # inlimit.
            self._inclause_limit = min(window_size, self._sql.default_inlimit)
            LOGGER.debug("SQL index: using default inclause limit.", tags=dict(inclause_limit=self._inclause_limit))

        # Make a test query against the database to ensure the connection is valid
        with self._sql.scoped_session() as session:
            session.query(IndexEntry).first()
        self._sql.remove_scoped_session()

    def start(self) -> None:
        self._storage.start()

    def stop(self) -> None:
        self._storage.stop()

    @timed(METRIC.STORAGE.STAT_DURATION, type=TYPE)
    def has_blob(self, digest: Digest) -> bool:
        with self._sql.scoped_session() as session:
            statement = select(func.count(IndexEntry.digest_hash)).where(IndexEntry.digest_hash == digest.hash)

            num_entries = session.execute(statement).scalar()
            if num_entries is None:
                num_entries = 0

            if num_entries == 1:
                return True
            elif num_entries < 1:
                return False
            else:
                raise RuntimeError(f"Multiple results found for blob [{digest}]. The index is in a bad state.")

    @timed(METRIC.STORAGE.READ_DURATION, type=TYPE)
    def get_blob(self, digest: Digest) -> Optional[IO[bytes]]:
        """Get a blob from the index or the backing storage. Optionally fallback and repair index"""

        # Check the index for the blob and return if found.
        with self._sql.scoped_session() as session:
            if entry := session.query(IndexEntry).filter(IndexEntry.digest_hash == digest.hash).first():
                if entry.inline_blob is not None:
                    return io.BytesIO(entry.inline_blob)
                elif blob := self._storage.get_blob(digest):
                    # Fix any blobs that should have been inlined.
                    if digest.size_bytes <= self._max_inline_blob_size:
                        self._save_digests_to_index([(digest, read_and_rewind(blob))], session)
                        session.commit()
                    return blob
                LOGGER.warning(
                    "Blob was indexed but not in storage. Deleting from the index.", tags=dict(digest=digest)
                )
                self._bulk_delete_from_index([digest], session)

        # Check the storage for the blob and repair the index if found.
        if self._fallback_on_get:
            if blob := self._storage.get_blob(digest):
                with self._sql.scoped_session() as session:
                    if digest.size_bytes <= self._max_inline_blob_size:
                        self._save_digests_to_index([(digest, read_and_rewind(blob))], session)
                    else:
                        self._save_digests_to_index([(digest, None)], session)
                    session.commit()
                return blob

        # Blob was not found in index or storage
        return None

    @timed(METRIC.STORAGE.DELETE_DURATION, type=TYPE)
    def delete_blob(self, digest: Digest) -> None:
        statement = delete(IndexEntry).where(IndexEntry.digest_hash == digest.hash)
        options = {"synchronize_session": False}

        with self._sql.scoped_session() as session:
            session.execute(statement, execution_options=options)

        self._storage.delete_blob(digest)

    @timed(METRIC.STORAGE.WRITE_DURATION, type=TYPE)
    def commit_write(self, digest: Digest, write_session: IO[bytes]) -> None:
        inline_blob = None
        if digest.size_bytes > self._max_inline_blob_size:
            self._storage.commit_write(digest, write_session)
        else:
            write_session.seek(0)
            inline_blob = write_session.read()
        try:
            with self._sql.scoped_session() as session:
                self._save_digests_to_index([(digest, inline_blob)], session)
        except DBAPIError as error:
            # Error has pgcode attribute (Postgres only)
            if hasattr(error.orig, "pgcode"):
                # imported here to avoid global dependency on psycopg2
                from psycopg2.errors import DiskFull, Error, OutOfMemory

                # 53100 == DiskFull && 53200 == OutOfMemory
                original_exception = cast(Error, error.orig)
                if isinstance(original_exception, (DiskFull, OutOfMemory)):
                    raise StorageFullError(
                        f"Postgres Error: {original_exception.pgerror} ({original_exception.pgcode}"
                    ) from error
            raise error

    def _partitioned_hashes(self, digests: Sequence[Digest]) -> Iterator[Iterator[str]]:
        """Given a long list of digests, split it into parts no larger than
        _inclause_limit and yield the hashes in each part.
        """
        for part_start in range(0, len(digests), self._inclause_limit):
            part_end = min(len(digests), part_start + self._inclause_limit)
            part_digests = itertools.islice(digests, part_start, part_end)
            yield map(lambda digest: digest.hash, part_digests)

    def _bulk_select_digests(
        self, digests: Sequence[Digest], fetch_blobs: bool = False, fetch_deleted: bool = True
    ) -> Iterator[IndexEntry]:
        """Generator that selects all rows matching a digest list.

        SQLAlchemy Core is used for this because the ORM has problems with
        large numbers of bind variables for WHERE IN clauses.

        We only select on the digest hash (not hash and size) to allow for
        index-only queries on db backends that support them.
        """
        index_table = IndexEntry.__table__
        with self._sql.scoped_session() as session:
            columns = [index_table.c.digest_hash]
            if fetch_blobs:
                columns.append(index_table.c.inline_blob)
            for part in self._partitioned_hashes(digests):
                stmt = select(*columns).where(index_table.c.digest_hash.in_(part))
                if not fetch_deleted:
                    stmt = stmt.where(not_(index_table.c.deleted))
                entries = session.execute(stmt)
                yield from entries  # type: ignore

    @timed(METRIC.STORAGE.SQL_INDEX.UPDATE_TIMESTAMP_DURATION)
    def _bulk_refresh_timestamps(
        self, digests: Sequence[Digest], session: SessionType, update_time: Optional[datetime] = None
    ) -> None:
        """Refresh all timestamps of the input digests.

        SQLAlchemy Core is used for this because the ORM is not suitable for
        bulk inserts and updates.

        https://docs.sqlalchemy.org/en/13/faq/performance.html#i-m-inserting-400-000-rows-with-the-orm-and-it-s-really-slow
        """
        # See discussion of __table__ typing in https://github.com/sqlalchemy/sqlalchemy/issues/9130
        index_table = cast(Table, IndexEntry.__table__)
        current_time = datetime.utcnow()

        # If a timestamp was passed in, use it. And always refreshes (no threshold).
        if update_time:
            timestamp = update_time
            last_accessed_threshold = current_time
        # Otherwise timestamp of digest will not refresh if it was last accessed more recent than this threshold.
        else:
            timestamp = current_time
            last_accessed_threshold = current_time - timedelta(seconds=self.refresh_accesstime_older_than)

        for part in self._partitioned_hashes(digests):
            # Generate the SQL Statement:
            #   UPDATE index SET accessed_timestamp=<timestamp>
            #   WHERE index.digest_hash IN
            #   (SELECT index.digest_hash FROM index
            #    WHERE index.digest_hash IN <part> AND WHERE index.accessed_timestamp < <last_accessed_threshold>
            #    FOR UPDATE SKIP LOCKED)
            stmt = (
                index_table.update()
                .where(
                    index_table.c.digest_hash.in_(
                        select(index_table.c.digest_hash)
                        .where(index_table.c.digest_hash.in_(part))
                        .where(index_table.c.accessed_timestamp < last_accessed_threshold)
                        .with_for_update(skip_locked=True)
                    )
                )
                .values(accessed_timestamp=timestamp)
            )
            session.execute(stmt)
            session.commit()

    @timed(METRIC.STORAGE.BULK_STAT_DURATION, type=TYPE)
    def missing_blobs(self, digests: List[Digest]) -> List[Digest]:
        # Blobs marked as deleted are considered as missing
        entries = self._bulk_select_digests(digests, fetch_deleted=False)

        found_hashes = {entry.digest_hash for entry in entries}

        # Split the digests into two found/missing lists
        found_digests, missing_digests = [], []
        for digest in digests:
            if digest.hash in found_hashes:
                found_digests.append(digest)
            else:
                missing_digests.append(digest)

        # Update all timestamps for blobs which were found
        with self._sql.scoped_session() as session:
            self._bulk_refresh_timestamps(found_digests, session)

        return missing_digests

    @timed(METRIC.STORAGE.SQL_INDEX.SAVE_DIGESTS_DURATION)
    def _save_digests_to_index(
        self, digest_blob_pairs: List[Tuple[Digest, Optional[bytes]]], session: SessionType
    ) -> None:
        """Helper to persist a list of digest/blob pairs to the index.

        Any digests present are updated, and new digests are inserted along with their inline blobs (if provided).
        Only blobs with size less than or equal to the max_inline_blob_size are inserted directly into the index.
        """
        if not digest_blob_pairs:
            return

        digest_blob_pairs = sorted(digest_blob_pairs, key=lambda pair: (pair[0].hash, pair[0].size_bytes))

        if self._dialect_delegate:
            try:
                self._dialect_delegate._save_digests_to_index(  # type: ignore
                    digest_blob_pairs, session, self._max_inline_blob_size
                )
                return
            except AttributeError:
                pass

        update_time = datetime.utcnow()
        # Figure out which digests we can just update
        digests = [digest for (digest, blob) in digest_blob_pairs]
        entries = self._bulk_select_digests(digests)
        # Map digests to new entries
        entries_not_present = {
            digest.hash: {
                "digest_hash": digest.hash,
                "digest_size_bytes": digest.size_bytes,
                "accessed_timestamp": update_time,
                "inline_blob": (blob if digest.size_bytes <= self._max_inline_blob_size else None),
                "deleted": False,
            }
            for (digest, blob) in digest_blob_pairs
        }

        entries_present = {}
        for entry in entries:
            entries_present[entry.digest_hash] = entries_not_present[entry.digest_hash]
            del entries_not_present[entry.digest_hash]

        if entries_not_present:
            session.bulk_insert_mappings(IndexEntry, entries_not_present.values())  # type: ignore
        if entries_present:
            session.bulk_update_mappings(IndexEntry, entries_present.values())  # type: ignore

    @timed(METRIC.STORAGE.BULK_WRITE_DURATION, type=TYPE)
    def bulk_update_blobs(  # pylint: disable=arguments-renamed
        self, digest_blob_pairs: List[Tuple[Digest, bytes]]
    ) -> List[Status]:
        """Implement the StorageABC's bulk_update_blobs method.

        The StorageABC interface takes in a list of digest/blob pairs and
        returns a list of results. The list of results MUST be ordered to
        correspond with the order of the input list."""
        pairs_to_store = []
        result_map = {}

        # For each blob, determine whether to store it in the backing storage or inline it
        for digest, blob in digest_blob_pairs:
            if validate_digest_data(digest, blob):
                if digest.size_bytes > self._max_inline_blob_size:
                    pairs_to_store.append((digest, blob))
                else:
                    result_map[digest.hash] = Status(code=code_pb2.OK)
            else:
                result_map[digest.hash] = Status(code=code_pb2.INVALID_ARGUMENT, message="Data doesn't match hash")
        missing_blobs = self.missing_blobs([digest for digest, _ in pairs_to_store])
        missing_blob_pairs = []
        for digest, blob in pairs_to_store:
            if digest not in missing_blobs:
                result_map[digest.hash] = Status(code=code_pb2.OK)
            else:
                missing_blob_pairs.append((digest, blob))

        backup_results = self._storage.bulk_update_blobs(missing_blob_pairs)

        for digest, result in zip(missing_blobs, backup_results):
            if digest.hash in result_map:
                # ERROR: blob was both inlined and backed up
                raise RuntimeError("Blob was both inlined and backed up.")
            result_map[digest.hash] = result

        # Generate the final list of results
        pairs_to_inline: List[Tuple[Digest, Optional[bytes]]] = []
        results = []
        for digest, blob in digest_blob_pairs:
            status = result_map.get(
                digest.hash,
                Status(code=code_pb2.UNKNOWN, message="SQL Index: unable to determine the status of this blob"),
            )
            results.append(status)
            if status.code == code_pb2.OK:
                pairs_to_inline.append((digest, blob))

        with self._sql.scoped_session() as session:
            self._save_digests_to_index(pairs_to_inline, session)

        return results

    def _bulk_read_blobs_with_fallback(self, digests: List[Digest]) -> Dict[str, bytes]:
        hash_to_digest: Dict[str, Digest] = {digest.hash: digest for digest in digests}
        results: Dict[str, bytes] = {}

        expected_storage_digests: List[Digest] = []
        # Fetch inlined blobs directly from the index
        entries = self._bulk_select_digests(digests, fetch_blobs=True)
        for e in entries:
            blob, digest_hash, digest = e.inline_blob, e.digest_hash, hash_to_digest[e.digest_hash]
            if blob is not None:
                results[digest_hash] = blob
                hash_to_digest.pop(digest_hash)
            else:
                # If a blob is not inlined then the blob is expected to be in storage
                expected_storage_digests.append(digest)

        # Fetch everything that wasn't inlined from the backing storage
        fetched_digests = self._storage.bulk_read_blobs(list(hash_to_digest.values()))

        # Save everything fetched from storage, inlining the blobs if they're small enough
        digest_pairs_to_save: List[Tuple[Digest, Optional[bytes]]] = []
        for digest_hash, blob_data in fetched_digests.items():
            if blob_data is not None:
                digest = hash_to_digest[digest_hash]
                if digest.size_bytes <= self._max_inline_blob_size:
                    digest_pairs_to_save.append((digest, blob_data))
                else:
                    digest_pairs_to_save.append((digest, None))
                results[digest_hash] = blob_data

        # List of digests found in storage
        acutal_storage_digest_hashes = set(
            digest_hash for (digest_hash, blob_data) in fetched_digests.items() if blob_data is not None
        )
        # Get a list of all the digests that were in the index but not found in storage
        digests_expected_not_in_storage: List[Digest] = []
        for expected_digest in expected_storage_digests:
            if expected_digest.hash not in acutal_storage_digest_hashes:
                LOGGER.warning(
                    "Blob was indexed but not in storage. Deleting from the index.", tags=dict(digest=digest)
                )
                digests_expected_not_in_storage.append(expected_digest)

        with self._sql.scoped_session() as session:
            self._save_digests_to_index(digest_pairs_to_save, session)
            if digests_expected_not_in_storage:
                self._bulk_delete_from_index(digests_expected_not_in_storage, session)
            session.commit()

        return results

    @timed(METRIC.STORAGE.BULK_READ_DURATION, type=TYPE)
    def bulk_read_blobs(self, digests: List[Digest]) -> Dict[str, bytes]:
        if self._fallback_on_get:
            return self._bulk_read_blobs_with_fallback(digests)

        # If fallback is disabled, query the index first and only
        # query the storage for blobs that weren't inlined there

        hash_to_digest = {digest.hash: digest for digest in digests}  # hash -> digest map
        results: Dict[str, bytes] = {}  # The final list of results (return value)
        digests_to_fetch: List[Digest] = []  # Digests that need to be fetched from storage
        digest_pairs_to_save: List[Tuple[Digest, Optional[bytes]]] = []  # Digests that need to be updated in the index

        # Fetch all of the digests in the database
        # Anything that wasn't already inlined needs to be fetched
        entries = self._bulk_select_digests(digests, fetch_blobs=True)
        for index_entry in entries:
            digest = hash_to_digest[index_entry.digest_hash]
            if index_entry.inline_blob is not None:
                results[index_entry.digest_hash] = index_entry.inline_blob
            else:
                digests_to_fetch.append(digest)

        # Caution: digest whose blob cannot be found from storage will be dropped.
        if digests_to_fetch:
            fetched_digests = self._storage.bulk_read_blobs(digests_to_fetch)
        else:
            fetched_digests = {}

        # Generate the list of inputs for _save_digests_to_index
        #
        # We only need to send blob data for small blobs fetched
        # from the storage since everything else is either too
        # big or already inlined
        for digest in digests_to_fetch:
            if blob_data := fetched_digests.get(digest.hash):
                if digest.size_bytes <= self._max_inline_blob_size:
                    digest_pairs_to_save.append((digest, blob_data))

        acutal_storage_digests = set(digest_hash for (digest_hash, _) in fetched_digests.items())
        # Get a list of all the digests that were in the index but not found in storage
        digests_expected_not_in_storage: List[Digest] = []
        for expected_digest in digests_to_fetch:
            if expected_digest.hash not in acutal_storage_digests:
                LOGGER.warning(
                    "Blob was indexed but not in storage. Deleting from the index.", tags=dict(digest=digest)
                )
                digests_expected_not_in_storage.append(expected_digest)

        # Update any blobs which need to be inlined
        with self._sql.scoped_session() as session:
            self._save_digests_to_index(digest_pairs_to_save, session)
            if digests_expected_not_in_storage:
                self._bulk_delete_from_index(digests_expected_not_in_storage, session)
            session.commit()

        results.update(fetched_digests)
        return results

    def _column_windows(
        self, session: SessionType, column: InstrumentedAttribute[Any]
    ) -> Iterator[ColumnElement[bool]]:
        """Adapted from the sqlalchemy WindowedRangeQuery recipe.
        https://github.com/sqlalchemy/sqlalchemy/wiki/WindowedRangeQuery

        This method breaks the timestamp range into windows and yields
        the borders of these windows to the callee. For example, the borders
        yielded by this might look something like
        ('2019-10-08 18:25:03.699863', '2019-10-08 18:25:03.751018')
        ('2019-10-08 18:25:03.751018', '2019-10-08 18:25:03.807867')
        ('2019-10-08 18:25:03.807867', '2019-10-08 18:25:03.862192')
        ('2019-10-08 18:25:03.862192',)

        _windowed_lru_digests uses these borders to form WHERE clauses for its
        SELECTs. In doing so, we make sure to repeatedly query the database for
        live updates, striking a balance between loading the entire resultset
        into memory and querying each row individually, both of which are
        inefficient in the context of a large index.

        The window size is a parameter and can be configured. A larger window
        size will yield better performance (fewer SQL queries) at the cost of
        memory (holding on to the results of the query) and accuracy (blobs
        may get updated while you're working on them), and vice versa for a
        smaller window size.
        """

        def int_for_range(start_id: Any, end_id: Any) -> ColumnElement[bool]:
            if end_id:
                return and_(column >= start_id, column < end_id)
            else:
                return column >= start_id  # type: ignore[no-any-return]

        # Constructs a query that:
        # 1. Gets all the timestamps in sorted order.
        # 2. Assign a row number to each entry.
        # 3. Only keep timestamps that are every other N row number apart. N="_all_blobs_window_size".
        # SELECT
        #   anon_1.index_accessed_timestamp AS anon_1_index_accessed_timestamp
        #   FROM (
        #       SELECT
        #       index.accessed_timestamp AS index_accessed_timestamp,
        #       row_number() OVER (ORDER BY index.accessed_timestamp) AS rownum
        #       FROM index
        #       )
        #  AS anon_1
        #  WHERE rownum % 1000=1
        #
        # Note:
        #  - This query can be slow due to checking each entry with "WHERE rownum % 1000=1".
        #  - These timestamps will be the basis for constructing the SQL "WHERE" clauses later.
        rownum = func.row_number().over(order_by=column).label("rownum")
        subq = select(column, rownum).subquery()

        # The upstream recipe noted in the docstring uses `subq.corresponding_column` here. That
        # method takes a KeyedColumnElement, which the ORM InstrumentedAttributes are not instances
        # of. Rather than switching to passing actual columns here, we can take advantage of controlling
        # the initial `select` to instead use the subquery columns directly and avoid ever calling this
        # method.
        #
        # See https://github.com/sqlalchemy/sqlalchemy/discussions/10325#discussioncomment-6952547.
        target_column = subq.columns[0]
        rownum_column = subq.columns[1]

        stmt = select(target_column)
        if self._all_blobs_window_size > 1:
            stmt = stmt.filter(rownum_column % self._all_blobs_window_size == 1)

        # Execute the underlying query against the database.
        # Ex: intervals = [t1, t1001, t2001, ...], q = [(t1, ), (t1001, ), (t2001, ), ...]
        intervals = list(session.scalars(stmt))

        # Generate the whereclauses
        while intervals:
            start = intervals.pop(0)
            if intervals:
                end = intervals[0]
            else:
                end = None
            # Ex: yield "WHERE IndexEntry.accessed_timestamp >= start AND IndexEntry.accessed_timestamp < end"
            yield int_for_range(start, end)

    def _windowed_lru_digests(
        self, q: "Query[Any]", column: InstrumentedAttribute[Any]
    ) -> Iterator[Tuple[IndexEntry, bool]]:
        """Generate a query for each window produced by _column_windows
        and yield the results one by one.
        """
        # Determine whether the conditions are met to make an SQL call to get new windows.
        msg = "Using stored LRU windows"
        if len(self._queue_of_whereclauses) == 0:
            msg = "Requesting new LRU windows."
            self._queue_of_whereclauses = deque(self._column_windows(q.session, column))
            self._delete_premarked_blobs = True

        msg += f" Number of windows remaining: {len(self._queue_of_whereclauses)}"
        LOGGER.debug(msg)

        while self._queue_of_whereclauses:
            whereclause = self._queue_of_whereclauses[0]
            window = q.filter(whereclause).order_by(column.asc())
            yield from window

            # If yield from window doesn't get to this point that means
            # the cleanup hasn't consumed all the content in a whereclause and exited.
            # Otherwise, the whereclause is exhausted and can be discarded.
            self._queue_of_whereclauses.popleft()

    def least_recent_digests(self) -> Iterator[Digest]:
        with self._sql.scoped_session() as session:
            # TODO: session.query is legacy, we should replace this with the `select` construct
            # as we do elsewhere.
            q = session.query(IndexEntry)
            for entry in self._windowed_lru_digests(q, IndexEntry.accessed_timestamp):
                # TODO make this generic or delete this method only used by tests.
                index_entry = cast(IndexEntry, entry)
                assert isinstance(index_entry.digest_hash, str)
                assert isinstance(index_entry.digest_size_bytes, int)
                yield Digest(hash=index_entry.digest_hash, size_bytes=index_entry.digest_size_bytes)

    @timed(METRIC.STORAGE.SQL_INDEX.SIZE_CALCULATION_DURATION)
    def get_total_size(self) -> int:
        statement = select(func.sum(IndexEntry.digest_size_bytes))
        with self._sql.scoped_session() as session:
            result = session.execute(statement).scalar()
            if result is None:
                result = 0
            return result

    @timed(METRIC.STORAGE.SQL_INDEX.DELETE_N_BYTES_DURATION)
    def delete_n_bytes(
        self, n_bytes: int, dry_run: bool = False, protect_blobs_after: Optional[datetime] = None
    ) -> int:
        """
        When using a SQL Index, entries with a delete marker are "in the process of being deleted".
        This is required because storage operations can't be safely tied to the SQL index transaction
        (one may fail independently of the other, and you end up inconsistent).

        The workflow is roughly as follows:
        - Start a SQL transaction.
        - Lock and mark the indexed items you want to delete.
        - Close the SQL transaction.
        - Perform the storage deletes
        - Start a SQL transaction.
        - Actually delete the index entries.
        - Close the SQL transaction.

        This means anything with deleted=False will always be present in the backing store. If it is marked
        deleted=True, and the process gets killed when deleting from the backing storage, only
        some of the items might actually be gone.

        The next time the cleaner starts up, it can try to do that delete again (ignoring 404s).
        Eventually that will succeed and the item will actually be removed from the DB. Only during
        the first run of batches do we consider already marked items. This avoids multiple cleanup
        daemons from competing with each other on every batch.
        """
        if protect_blobs_after is None:
            protect_blobs_after = datetime.utcnow()

        # Used for metric publishing
        delete_start_time = time.time()

        storage_digests: List[Digest] = []
        marked_digests: List[Digest] = []
        collected_bytes = 0

        with self._sql.scoped_session(exceptions_to_not_rollback_on=[StaleDataError]) as session:
            base_query = session.query(IndexEntry, IndexEntry.inline_blob != None).options(  # noqa
                load_only(IndexEntry.digest_hash, IndexEntry.digest_size_bytes)
            )

            if self._delete_premarked_blobs:
                LOGGER.info("Starting to gather pre-marked records.")
                premarked_query = base_query.filter_by(deleted=True)
                for [entry, is_inline] in premarked_query.all():
                    digest = Digest(hash=entry.digest_hash, size_bytes=entry.digest_size_bytes)
                    marked_digests.append(digest)
                    if not is_inline:
                        storage_digests.append(digest)
                    collected_bytes += entry.digest_size_bytes

                if not dry_run:
                    publish_counter_metric(METRIC.STORAGE.SQL_INDEX.PREMARKED_DELETED_COUNT, len(marked_digests))
                LOGGER.info(
                    "Gathered pre-marked bytes.", tags=dict(collected_bytes=collected_bytes, max_bytes=n_bytes)
                )
                self._delete_premarked_blobs = False

            if collected_bytes < n_bytes:
                LOGGER.info("Searching for records to mark deleted.")
                unmarked_query = (
                    base_query.filter_by(deleted=False)
                    .filter(IndexEntry.accessed_timestamp < protect_blobs_after)
                    .with_for_update(skip_locked=True)
                )
                window = self._windowed_lru_digests(unmarked_query, IndexEntry.accessed_timestamp)
                mark_deleted_start = time.perf_counter()
                for [entry, is_inline] in window:
                    digest = Digest(hash=entry.digest_hash, size_bytes=entry.digest_size_bytes)
                    marked_digests.append(digest)
                    if not is_inline:
                        storage_digests.append(digest)
                    collected_bytes += entry.digest_size_bytes
                    if not dry_run:
                        entry.deleted = True
                    if collected_bytes >= n_bytes:
                        break
                mark_deleted_duration = timedelta(seconds=time.perf_counter() - mark_deleted_start)
                if not dry_run:
                    publish_timer_metric(METRIC.STORAGE.SQL_INDEX.MARK_DELETED_DURATION, mark_deleted_duration)
                LOGGER.info("Gathered bytes.", tags=dict(collected_bytes=collected_bytes, max_bytes=n_bytes))

        if dry_run:
            return collected_bytes

        failed_deletes = self._storage.bulk_delete(storage_digests)
        digests_to_delete = [x for x in marked_digests if f"{x.hash}/{x.size_bytes}" not in failed_deletes]

        with self._sql.scoped_session(exceptions_to_not_rollback_on=[StaleDataError]) as session:
            failed_deletes.extend(self._bulk_delete_from_index(digests_to_delete, session))
        for digest in digests_to_delete:
            if digest in failed_deletes:
                collected_bytes -= digest.size_bytes

        batch_duration = time.time() - delete_start_time
        blobs_deleted_per_second = (len(digests_to_delete) - len(failed_deletes)) / batch_duration
        publish_gauge_metric(METRIC.CLEANUP.BLOBS_DELETED_PER_SECOND, blobs_deleted_per_second)
        return collected_bytes

    @timed(METRIC.STORAGE.BULK_DELETE_DURATION, type=TYPE)
    def bulk_delete(self, digests: List[Digest]) -> List[str]:
        # Delete from the index and then delete from the backing storage.
        with self._sql.scoped_session(exceptions_to_not_rollback_on=[StaleDataError]) as session:
            failed_deletes = self._bulk_delete_from_index(digests, session)

        digests_to_delete = [x for x in digests if f"{x.hash}/{x.size_bytes}" not in failed_deletes]
        failed_deletes.extend(self._storage.bulk_delete(digests_to_delete))
        return failed_deletes

    @timed(METRIC.STORAGE.SQL_INDEX.BULK_DELETE_INDEX_DURATION)
    def _bulk_delete_from_index(self, digests: List[Digest], session: Session) -> List[str]:
        LOGGER.info("Deleting digests from the index.", tags=dict(digest_count=len(digests)))
        # See discussion of __table__ typing in https://github.com/sqlalchemy/sqlalchemy/issues/9130
        index_table = cast(Table, IndexEntry.__table__)
        hashes = [x.hash for x in digests]

        # Make sure we don't exceed maximum size of an IN clause
        n = self._inclause_limit
        hash_chunks = [hashes[i : i + n] for i in range(0, len(hashes), n)]

        # We will not raise, rollback, or log on StaleDataErrors.
        # These errors occur when we delete fewer rows than we were expecting.
        # This is fine, since the missing rows will get deleted eventually.
        # When running bulk_deletes concurrently, StaleDataErrors
        # occur too often to log.
        num_blobs_deleted = 0
        for chunk in hash_chunks:
            # Do not wait for locks when deleting rows. Skip locked rows to
            # avoid deadlocks.
            stmt = index_table.delete().where(
                index_table.c.digest_hash.in_(
                    select(index_table.c.digest_hash)
                    .where(index_table.c.digest_hash.in_(chunk))
                    .with_for_update(skip_locked=True)
                )
            )
            num_blobs_deleted += session.execute(stmt).rowcount
        LOGGER.info("Blobs deleted from the index.", tags=dict(deleted_count=num_blobs_deleted, digest_count=digests))

        # bulk_delete is typically expected to return the digests that were not deleted,
        # but delete only returns the number of rows deleted and not what was/wasn't
        # deleted. Getting this info would require extra queries, so assume that
        # everything was either deleted or already deleted. Failures will continue to throw
        return []



# Copyright (C) 2018 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
CommandLineInterface
===================

Any files in the commands/ folder with the name cmd_*.py
will be attempted to be imported.
"""

import importlib
import logging
import os
import sys
from typing import Any, Dict, List, Optional, cast

import click
import grpc
from importlib_resources import files

from buildgrid.server.metadata import ctx_grpc_request_id
from buildgrid.server.settings import LOG_RECORD_FORMAT

CONTEXT_SETTINGS = {
    "auto_envvar_prefix": "BUILDGRID",
    "max_content_width": 1000,  # do not truncate option help unnecessarily
}


class Context:
    def __init__(self) -> None:
        self.verbose: bool = False
        self.channel: Optional[grpc.Channel] = None
        self.cas_channel: Optional[grpc.Channel] = None
        self.operations_channel: Optional[grpc.Channel] = None
        self.cache_channel: Optional[grpc.Channel] = None
        self.logstream_channel: Optional[grpc.Channel] = None
        self.instance_name: str = ""

        self.user_home: str = os.getcwd()


pass_context = click.make_pass_decorator(Context, ensure=True)
cmd_folder = files("buildgrid.server.app").joinpath("commands")

commands_namespace = "buildgrid.server.app.commands"


class App(click.MultiCommand):
    def list_commands(self, ctx: Any) -> List[str]:
        """Lists available command names."""
        commands = []
        for filename in os.listdir(cmd_folder):
            if filename.endswith(".py") and filename.startswith("cmd_"):
                command_name = filename[4:-3].replace("_", "-")
                commands.append(command_name)
        commands.sort()

        return commands

    def get_command(self, ctx: Any, cmd_name: str) -> click.Command:
        """Looks-up and loads a particular command by name."""
        try:
            module_name = cmd_name.replace("-", "_")
            module = importlib.import_module(f"{commands_namespace}.cmd_{module_name}")
        except ImportError as e:
            failed_import_name = e.name
            assert failed_import_name is not None
            if failed_import_name.startswith(commands_namespace):
                click.echo(f"Error: No such command: [{cmd_name}].", err=True)
                sys.exit(-1)
            else:
                return click.Command(cmd_name, help=f"[unavailable, module '{e.name}' is not installed]")

        return cast(click.Command, module.cli)


class DebugFilter(logging.Filter):
    def __init__(self, debug_domains: str, name: str = "") -> None:
        super().__init__(name=name)
        # Recursive dict. RDict = Dict[str, RDict]
        self.__domains_tree: Dict[str, Any] = {}

        for domain in debug_domains.split(":"):
            domains_tree = self.__domains_tree
            for label in domain.split("."):
                if all(key not in domains_tree for key in [label, "*"]):
                    domains_tree[label] = {}
                domains_tree = domains_tree[label]

    def filter(self, record: logging.LogRecord) -> bool:
        # Only evaluate DEBUG records for filtering
        if record.levelname != "DEBUG":
            return True
        domains_tree, last_match = self.__domains_tree, None
        for label in record.name.split("."):
            if all(key not in domains_tree for key in [label, "*"]):
                return False
            last_match = label if label in domains_tree else "*"
            domains_tree = domains_tree[last_match]
        if domains_tree and "*" not in domains_tree:
            return False
        return True


class RequestIdFilter(logging.Filter):
    def filter(self, record: logging.LogRecord) -> bool:
        record.request_id = ctx_grpc_request_id.get() or "---"
        return True


def setup_logging(verbosity: int = 0, debug_mode: bool = False) -> None:
    """Deals with loggers verbosity"""
    asyncio_logger = logging.getLogger("asyncio")
    root_logger = logging.getLogger()

    log_handler = logging.StreamHandler(stream=sys.stdout)
    for log_filter in root_logger.filters:
        log_handler.addFilter(log_filter)

    log_handler.addFilter(RequestIdFilter())

    logging.basicConfig(format=LOG_RECORD_FORMAT, handlers=[log_handler])

    if verbosity == 1:
        root_logger.setLevel(logging.WARNING)
    elif verbosity == 2:
        root_logger.setLevel(logging.INFO)
    elif verbosity >= 3:
        root_logger.setLevel(logging.DEBUG)
    else:
        root_logger.setLevel(logging.ERROR)

    if not debug_mode:
        asyncio_logger.setLevel(logging.CRITICAL)
    else:
        asyncio_logger.setLevel(logging.DEBUG)
        root_logger.setLevel(logging.DEBUG)


@click.command(cls=App, context_settings=CONTEXT_SETTINGS)
@pass_context
def cli(context: Any) -> None:
    """BuildGrid's client and server CLI front-end."""
    root_logger = logging.getLogger()

    # Clean-up root logger for any pre-configuration:
    for log_handler in root_logger.handlers[:]:
        root_logger.removeHandler(log_handler)
    for log_filter in root_logger.filters[:]:
        root_logger.removeFilter(log_filter)

    # Filter debug messages using BGD_MESSAGE_DEBUG value:
    debug_domains = os.environ.get("BGD_MESSAGE_DEBUG", None)
    if debug_domains:
        root_logger.addFilter(DebugFilter(debug_domains))



# Copyright (C) 2024 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
from typing import Any, Optional, Union

import yaml
from pydantic import BaseModel, Field

from buildgrid.server.app.settings.parser import string_definitions


class S3Config(BaseModel):
    access_key: str
    bucket_regex: str
    endpoint: str
    path_prefix: str
    hash_prefix_size: int = Field(default=0)
    secret_key: str


class RedisConfig(BaseModel):
    db: Optional[int] = Field(default=None)
    dns_srv_record: Optional[str] = Field(default=None)
    index_prefix: str
    key_batch_size: int
    password: Optional[str] = Field(default=None)
    host: Optional[str] = Field(default=None)
    port: Optional[int] = Field(default=None)
    sentinel_master_name: Optional[str] = Field(default=None)


class JanitorConfig(BaseModel):
    redis: Optional[RedisConfig] = Field(default=None)
    sleep_interval: int
    s3: S3Config
    sql_connection_string: Optional[str] = Field(default=None)


def parse_janitor_config(path: Union[str, bytes, "os.PathLike[str]"]) -> JanitorConfig:
    class Loader(yaml.SafeLoader):
        def string_loader(self, node: yaml.MappingNode) -> Any:
            return string_definitions[node.tag](node.value)

    for kind in string_definitions:
        Loader.add_constructor(kind, Loader.string_loader)

    with open(path) as config_file:
        config = yaml.load(config_file, Loader=Loader)
    return JanitorConfig(**config)



# Copyright (C) 2024 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Add property label

Revision ID: 23f48374c8b6
Revises: fe3a61461feb
Create Date: 2024-09-05 11:49:28.973640

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "23f48374c8b6"
down_revision = "fe3a61461feb"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("jobs", sa.Column("property_label", sa.String(), server_default="unknown", nullable=False))
    op.drop_index("ix_jobs_stage", table_name="jobs")
    op.create_index("ix_jobs_stage_property_label", "jobs", ["stage", "property_label"], unique=False)
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("ix_jobs_stage_property_label", table_name="jobs")
    op.create_index("ix_jobs_stage", "jobs", ["stage"], unique=False)
    op.drop_column("jobs", "property_label")
    # ### end Alembic commands ###



# Copyright (C) 2018 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

import pytest
from buildgrid_metering.client.client import SyncMeteringServiceClient
from buildgrid_metering.client.exceptions import MeteringServiceClientError
from buildgrid_metering.models.api import GetThrottlingResponse
from buildgrid_metering.models.dataclasses import ComputingUsage, Identity, RPCUsage, Usage
from grpc._server import _Context

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import Action, Command, Digest
from buildgrid._protos.google.longrunning import operations_pb2
from buildgrid.server.enums import MeteringThrottleAction
from buildgrid.server.exceptions import CancelledError, FailedPreconditionError, ResourceExhaustedError
from buildgrid.server.execution.instance import ExecutionInstance
from buildgrid.server.scheduler import DynamicPropertySet
from buildgrid.server.sql.models import ClientIdentityEntry


class MockDataStore:
    def __init__(self, storage):
        self.storage = storage


class MockScheduler:
    def __init__(self, storage, property_set, metering_client=None, metering_throttle_action=None):
        self.storage = storage
        self.property_set = property_set
        self.metering_client = metering_client
        self.metering_throttle_action = metering_throttle_action or MeteringThrottleAction.DEPRIORITIZE

        self._kwargs = None
        self._args = None

    def queue_job_action(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs
        return "queued"

    def return_queue_call_args(self):
        return self._args, self._kwargs


class DeprecatedAction:
    def __init__(self, *, command_digest, input_root_digest, do_not_cache):
        self.command_digest = command_digest
        self.input_root_digest = input_root_digest
        self.do_not_cache = do_not_cache

    def HasField(self, field):
        return False


class MockStorage:
    def __init__(self, pairs):
        command_digest = Digest(hash="command_digest", size_bytes=0)
        input_root_digest = Digest(hash="input_root_digest", size_bytes=0)
        self.action = Action(command_digest=command_digest, input_root_digest=input_root_digest, do_not_cache=True)
        self.command = Command()
        for property_pair in pairs:
            prop = self.command.platform.properties.add()
            prop.name = property_pair[0]
            prop.value = property_pair[1]

            prop = self.action.platform.properties.add()
            prop.name = property_pair[0]
            prop.value = property_pair[1]

        self.action_no_platform = DeprecatedAction(
            command_digest=command_digest, input_root_digest=input_root_digest, do_not_cache=True
        )
        self.command_fetched = False

    def get_message(self, digest, *args, **kwargs):
        if digest == "fake":
            return self.action
        elif digest == "fake-no-platform":
            return self.action_no_platform
        else:
            self.command_fetched = True
            return self.command


def test_execute_platform_matching_simple():
    """Will match on standard keys."""

    pairs = [("OSFamily", "linux"), ("ISA", "x86-64"), ("ISA", "x86-avx")]
    storage = MockStorage(pairs)
    property_set = DynamicPropertySet(
        unique_property_keys={"OSFamily"},
        match_property_keys={"OSFamily", "ISA"},
        wildcard_property_keys=set(),
    )
    exec_instance = ExecutionInstance(MockScheduler(storage, property_set=property_set))
    assert exec_instance.execute(action_digest="fake", skip_cache_lookup=False) == "queued"


def test_execute_platform_matching_fallback():
    """Will match on standard keys."""

    pairs = [("OSFamily", "linux"), ("ISA", "x86-64"), ("ISA", "x86-avx")]
    storage = MockStorage(pairs)
    property_set = DynamicPropertySet(
        unique_property_keys={"OSFamily"},
        match_property_keys={"OSFamily", "ISA"},
        wildcard_property_keys=set(),
    )
    exec_instance = ExecutionInstance(MockScheduler(storage, property_set=property_set))
    assert exec_instance.execute(action_digest="fake-no-platform", skip_cache_lookup=False) == "queued"
    assert storage.command_fetched


def test_execute_platform_matching_too_many_os():
    """Will not match due to too many OSFamilies being specified."""

    pairs = [("OSFamily", "linux"), ("OSFamily", "macos"), ("ISA", "x86-64"), ("ISA", "x86-avx")]
    property_set = DynamicPropertySet(
        unique_property_keys={"OSFamily"},
        match_property_keys={"OSFamily", "ISA"},
        wildcard_property_keys=set(),
    )
    exec_instance = ExecutionInstance(MockScheduler(MockStorage(pairs), property_set=property_set))
    with pytest.raises(FailedPreconditionError):
        exec_instance.execute(action_digest="fake", skip_cache_lookup=False)


def test_execute_platform_matching_too_many_os_platform_key():
    """Make sure adding duplicate keys won't cause issues ."""

    pairs = [("OSFamily", "linux"), ("OSFamily", "macos"), ("ISA", "x86-64"), ("ISA", "x86-avx")]
    property_set = DynamicPropertySet(
        unique_property_keys={"OSFamily"},
        match_property_keys={"OSFamily", "ISA"},
        wildcard_property_keys=set(),
    )
    exec_instance = ExecutionInstance(MockScheduler(MockStorage(pairs), property_set=property_set))
    with pytest.raises(FailedPreconditionError):
        exec_instance.execute(action_digest="fake", skip_cache_lookup=False)


def test_execute_platform_matching_failure():
    """Will not match due to platform-keys missing 'ChrootDigest'."""

    pairs = [("OSFamily", "linux"), ("ISA", "x86-64"), ("ChrootDigest", "deadbeef")]
    property_set = DynamicPropertySet(
        unique_property_keys={"OSFamily"},
        match_property_keys={"OSFamily", "ISA"},
        wildcard_property_keys=set(),
    )
    exec_instance = ExecutionInstance(MockScheduler(MockStorage(pairs), property_set=property_set))
    with pytest.raises(FailedPreconditionError):
        exec_instance.execute(action_digest="fake", skip_cache_lookup=False)


def test_execute_platform_matching_success():
    """Will match due to platform keys matching."""

    pairs = [("ChrootDigest", "deadbeed")]
    property_set = DynamicPropertySet(
        unique_property_keys={"OSFamily"},
        match_property_keys={"OSFamily", "ISA"},
        wildcard_property_keys={"ChrootDigest"},
    )
    exec_instance = ExecutionInstance(MockScheduler(MockStorage(pairs), property_set=property_set))
    assert exec_instance.execute(action_digest="fake", skip_cache_lookup=False) == "queued"


def test_execute_platform_matching_config_only():
    """Will match due to platform keys matching."""

    pairs = [("OSFamily", "linux"), ("ISA", "x86-64"), ("ChrootDigest", "deadbeed")]
    property_set = DynamicPropertySet(
        unique_property_keys={"OSFamily"},
        match_property_keys={"OSFamily", "ISA"},
        wildcard_property_keys={"ChrootDigest"},
    )
    scheduler = MockScheduler(MockStorage(pairs), property_set=property_set)
    exec_instance = ExecutionInstance(scheduler)
    assert exec_instance.execute(action_digest="fake", skip_cache_lookup=False) == "queued"
    # The ChrootDigest key shouldn't make it into the actual requirements
    _, queue_kwargs = scheduler.return_queue_call_args()
    assert "ChrootDigest" not in queue_kwargs["platform_requirements"]
    assert all(key in queue_kwargs["platform_requirements"] for key in property_set.match_property_keys)


def test_execute_platform_matching_both_empty():
    """Edge case where nothing specified on either side."""

    pairs = []
    property_set = DynamicPropertySet(
        unique_property_keys=set(),
        match_property_keys=set(),
        wildcard_property_keys=set(),
    )
    exec_instance = ExecutionInstance(MockScheduler(MockStorage(pairs), property_set=property_set))
    assert exec_instance.execute(action_digest="fake", skip_cache_lookup=False) == "queued"


def test_execute_platform_matching_no_job_req():
    """If job doesn't specify platform key requirements, it should always pass."""

    pairs = [("OSFamily", "linux"), ("ISA", "x86-64")]
    property_set = DynamicPropertySet(
        unique_property_keys={"OSFamily"},
        match_property_keys={"OSFamily", "ISA"},
        wildcard_property_keys={"ChrootDigest"},
    )
    exec_instance = ExecutionInstance(MockScheduler(MockStorage(pairs), property_set=property_set))
    assert exec_instance.execute(action_digest="fake", skip_cache_lookup=False) == "queued"


def test_execute_priority_set():
    """Check that the priority gets set."""
    pairs = [("OSFamily", "linux"), ("ISA", "x86-64")]
    property_set = DynamicPropertySet(
        unique_property_keys={"OSFamily"},
        match_property_keys={"OSFamily", "ISA"},
        wildcard_property_keys={"ChrootDigest"},
    )
    mock_scheduler = MockScheduler(MockStorage(pairs), property_set=property_set)
    exec_instance = ExecutionInstance(mock_scheduler)
    exec_instance.execute(action_digest="fake", skip_cache_lookup=False, priority=3)
    _, kwargs = mock_scheduler.return_queue_call_args()
    assert "priority" in kwargs
    assert kwargs["priority"] == 3


def test_execute_priority_default():
    """Check that the priority gets sets to 0 when not specified."""
    pairs = [("OSFamily", "linux"), ("ISA", "x86-64")]
    property_set = DynamicPropertySet(
        unique_property_keys={"OSFamily"},
        match_property_keys={"OSFamily", "ISA"},
        wildcard_property_keys={"ChrootDigest"},
    )
    mock_scheduler = MockScheduler(MockStorage(pairs), property_set=property_set)
    exec_instance = ExecutionInstance(mock_scheduler)
    exec_instance.execute(action_digest="fake", skip_cache_lookup=False)
    _, kwargs = mock_scheduler.return_queue_call_args()
    assert "priority" in kwargs
    assert kwargs["priority"] == 0


@pytest.fixture
def mock_exec_instance():
    return ExecutionInstance(MockScheduler(MockStorage([]), [], []))


@pytest.fixture
def mock_active_context():
    cxt = mock.MagicMock(spec=_Context)
    yield cxt


@pytest.fixture
def mock_operation():
    operation = mock.Mock(spec=operations_pb2.Operation)
    operation.done = False
    return operation


@pytest.fixture
def mock_operation_done():
    operation = mock.Mock(spec=operations_pb2.Operation)
    operation.done = True
    return operation


@pytest.fixture(params=[0, 1, 2])
def operation_updates_message_seq(mock_operation, request):
    seq = []
    for i in range(request.param):
        seq.append((None, mock_operation))

    return (request.param, seq)


@pytest.fixture(params=[0, 1, 2])
def operation_updates_completing_message_seq(mock_operation, mock_operation_done, request):
    seq = []
    for i in range(request.param):
        seq.append((None, mock_operation))

    seq.append((None, mock_operation_done))

    # Add sentinel operation update: this should never be dequeued
    # since we should stop once the operation completed (above)
    seq.append((ValueError, mock_operation))
    return (request.param, seq)


@pytest.fixture(params=[0, 1, 2])
def operation_updates_ending_with_error_seq(mock_operation, mock_operation_done, request):
    seq = []
    for i in range(request.param):
        seq.append((None, mock_operation))

    seq.append((CancelledError("Operation has been cancelled"), mock_operation))

    # Add sentinel operation update: this should never be dequeued
    # since we should stop once we encounter the first error
    seq.append((ValueError, mock_operation))
    return (request.param, seq)


@pytest.fixture(params=[0, 1, 2, 3])
def n_0_to_3_inclusive(request):
    return request.param


@pytest.fixture
def client_identity():
    return ClientIdentityEntry(instance="", workflow="build", actor="tool", subject="user")


@pytest.fixture
def mock_metering_client() -> SyncMeteringServiceClient:
    mock_metering_client = mock.Mock()
    mock_metering_client.get_throttling.return_value = GetThrottlingResponse(throttled=False)
    mock_metering_client.put_usage.return_value = None
    return mock_metering_client


def test_execute_not_throttled(client_identity: ClientIdentityEntry, mock_metering_client: SyncMeteringServiceClient):
    pairs = [("OSFamily", "linux"), ("ISA", "x86-64")]
    property_set = DynamicPropertySet(
        unique_property_keys={"OSFamily"},
        match_property_keys={"OSFamily", "ISA"},
        wildcard_property_keys={"ChrootDigest"},
    )
    mock_scheduler = MockScheduler(MockStorage(pairs), property_set=property_set, metering_client=mock_metering_client)
    exec_instance = ExecutionInstance(mock_scheduler)

    exec_instance.execute(action_digest="fake", skip_cache_lookup=False, priority=-1, client_identity=client_identity)

    _, kwargs = mock_scheduler.return_queue_call_args()
    assert "priority" in kwargs
    assert kwargs["priority"] == -1
    mock_metering_client.put_usage.assert_has_calls(
        [
            mock.call(
                Identity(
                    instance=client_identity.instance,
                    workflow=client_identity.workflow,
                    actor=client_identity.actor,
                    subject=client_identity.subject,
                ),
                "queued",
                Usage(rpc=RPCUsage(execute=1)),
            )
        ]
    )


def test_execute_not_throttled_given_exception(
    client_identity: ClientIdentityEntry, mock_metering_client: SyncMeteringServiceClient
):
    pairs = [("OSFamily", "linux"), ("ISA", "x86-64")]
    property_set = DynamicPropertySet(
        unique_property_keys={"OSFamily"},
        match_property_keys={"OSFamily", "ISA"},
        wildcard_property_keys={"ChrootDigest"},
    )
    mock_metering_client.get_throttling.side_effect = MeteringServiceClientError("connection reset")
    mock_scheduler = MockScheduler(MockStorage(pairs), property_set=property_set, metering_client=mock_metering_client)
    exec_instance = ExecutionInstance(mock_scheduler)

    exec_instance.execute(action_digest="fake", skip_cache_lookup=False, priority=-1, client_identity=client_identity)

    _, kwargs = mock_scheduler.return_queue_call_args()
    assert "priority" in kwargs
    assert kwargs["priority"] == -1


def test_execute_not_throttled_if_no_client_id(mock_metering_client: SyncMeteringServiceClient):
    pairs = [("OSFamily", "linux"), ("ISA", "x86-64")]
    property_set = DynamicPropertySet(
        unique_property_keys={"OSFamily"},
        match_property_keys={"OSFamily", "ISA"},
        wildcard_property_keys={"ChrootDigest"},
    )
    mock_scheduler = MockScheduler(MockStorage(pairs), property_set=property_set, metering_client=mock_metering_client)
    exec_instance = ExecutionInstance(mock_scheduler)

    exec_instance.execute(action_digest="fake", skip_cache_lookup=False, priority=-1, client_identity=None)

    _, kwargs = mock_scheduler.return_queue_call_args()
    assert "priority" in kwargs
    assert kwargs["priority"] == -1
    assert mock_metering_client.put_usage.call_count == 0


def test_execute_not_throttled_if_no_client(client_identity: ClientIdentityEntry):
    pairs = [("OSFamily", "linux"), ("ISA", "x86-64")]
    property_set = DynamicPropertySet(
        unique_property_keys={"OSFamily"},
        match_property_keys={"OSFamily", "ISA"},
        wildcard_property_keys={"ChrootDigest"},
    )
    mock_scheduler = MockScheduler(MockStorage(pairs), property_set=property_set, metering_client=None)
    exec_instance = ExecutionInstance(mock_scheduler)

    exec_instance.execute(action_digest="fake", skip_cache_lookup=False, priority=-1, client_identity=client_identity)

    _, kwargs = mock_scheduler.return_queue_call_args()
    assert "priority" in kwargs
    assert kwargs["priority"] == -1


def test_execute_throttled(client_identity: ClientIdentityEntry, mock_metering_client: SyncMeteringServiceClient):
    pairs = [("OSFamily", "linux"), ("ISA", "x86-64")]
    property_set = DynamicPropertySet(
        unique_property_keys={"OSFamily"},
        match_property_keys={"OSFamily", "ISA"},
        wildcard_property_keys={"ChrootDigest"},
    )
    mock_metering_client.get_throttling.return_value = GetThrottlingResponse(
        throttled=True, tracked_time_window_secs=100, tracked_usage=Usage(computing=ComputingUsage(utime=1))
    )
    mock_scheduler = MockScheduler(MockStorage(pairs), property_set=property_set, metering_client=mock_metering_client)
    exec_instance = ExecutionInstance(mock_scheduler)

    exec_instance.execute(action_digest="fake", skip_cache_lookup=False, priority=-1, client_identity=client_identity)

    _, kwargs = mock_scheduler.return_queue_call_args()
    assert "priority" in kwargs
    assert kwargs["priority"] == 1
    mock_metering_client.put_usage.assert_has_calls(
        [
            mock.call(
                Identity(
                    instance=client_identity.instance,
                    workflow=client_identity.workflow,
                    actor=client_identity.actor,
                    subject=client_identity.subject,
                ),
                "queued",
                Usage(rpc=RPCUsage(execute=1)),
            )
        ]
    )


def test_execute_throttled_rejected(
    client_identity: ClientIdentityEntry, mock_metering_client: SyncMeteringServiceClient
):
    pairs = [("OSFamily", "linux"), ("ISA", "x86-64")]
    property_set = DynamicPropertySet(
        unique_property_keys={"OSFamily"},
        match_property_keys={"OSFamily", "ISA"},
        wildcard_property_keys={"ChrootDigest"},
    )
    mock_metering_client.get_throttling.return_value = GetThrottlingResponse(
        throttled=True, tracked_time_window_secs=100, tracked_usage=Usage(computing=ComputingUsage(utime=1))
    )
    mock_scheduler = MockScheduler(
        MockStorage(pairs),
        property_set=property_set,
        metering_client=mock_metering_client,
        metering_throttle_action=MeteringThrottleAction.REJECT,
    )
    exec_instance = ExecutionInstance(mock_scheduler)

    with pytest.raises(ResourceExhaustedError):
        exec_instance.execute(
            action_digest="fake", skip_cache_lookup=False, priority=-1, client_identity=client_identity
        )



# Copyright (C) 2019 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
from typing import List, Optional

from sqlalchemy import BigInteger, Column, ForeignKey, Index, Table, UniqueConstraint, false
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, registry, relationship
from typing_extensions import Annotated

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import Action, Digest
from buildgrid._protos.google.devtools.remoteworkers.v1test2 import bots_pb2
from buildgrid.server.enums import LeaseState

bigint = Annotated[int, "bigint"]


class Base(DeclarativeBase):
    registry = registry(type_annotation_map={bigint: BigInteger()})


job_platform_association = Table(
    "job_platforms",
    Base.metadata,
    Column("job_name", ForeignKey("jobs.name", ondelete="CASCADE", onupdate="CASCADE"), primary_key=True),
    Column("platform_id", ForeignKey("platform_properties.id"), primary_key=True),
)


class PlatformEntry(Base):
    __tablename__ = "platform_properties"
    __table_args__ = (UniqueConstraint("key", "value"),)

    id: Mapped[int] = mapped_column(autoincrement=True, primary_key=True)
    key: Mapped[str]
    value: Mapped[str]

    jobs: Mapped[List["JobEntry"]] = relationship(
        "JobEntry", secondary=job_platform_association, back_populates="platform"
    )


class JobEntry(Base):
    __tablename__ = "jobs"

    # Immutable data
    name: Mapped[str] = mapped_column(primary_key=True)
    instance_name: Mapped[str] = mapped_column(index=True)
    action_digest: Mapped[str] = mapped_column(index=True)
    action: Mapped[bytes]
    do_not_cache: Mapped[bool] = mapped_column(default=False)
    # This is a hash of the platform properties, used for matching jobs to workers
    platform_requirements: Mapped[str]
    property_label: Mapped[str] = mapped_column(server_default="unknown")
    command: Mapped[str]

    # Scheduling state
    stage: Mapped[int] = mapped_column(default=0)
    priority: Mapped[int] = mapped_column(default=1, index=True)
    cancelled: Mapped[bool] = mapped_column(default=False)
    assigned: Mapped[bool] = mapped_column(default=False)
    n_tries: Mapped[int] = mapped_column(default=0)

    # Return data on completion
    result: Mapped[Optional[str]]
    status_code: Mapped[Optional[int]]

    # Auditing data
    create_timestamp: Mapped[Optional[datetime.datetime]]
    queued_timestamp: Mapped[datetime.datetime] = mapped_column(index=True)
    queued_time_duration: Mapped[Optional[int]]
    worker_start_timestamp: Mapped[Optional[datetime.datetime]]
    worker_completed_timestamp: Mapped[Optional[datetime.datetime]]
    input_fetch_start_timestamp: Mapped[Optional[datetime.datetime]]
    input_fetch_completed_timestamp: Mapped[Optional[datetime.datetime]]
    output_upload_start_timestamp: Mapped[Optional[datetime.datetime]]
    output_upload_completed_timestamp: Mapped[Optional[datetime.datetime]]
    execution_start_timestamp: Mapped[Optional[datetime.datetime]]
    execution_completed_timestamp: Mapped[Optional[datetime.datetime]]

    # Logstream identifiers
    stdout_stream_name: Mapped[Optional[str]]
    stdout_stream_write_name: Mapped[Optional[str]]
    stderr_stream_name: Mapped[Optional[str]]
    stderr_stream_write_name: Mapped[Optional[str]]

    leases: Mapped[List["LeaseEntry"]] = relationship("LeaseEntry", back_populates="job")
    active_states: List[int] = [
        LeaseState.UNSPECIFIED.value,
        LeaseState.PENDING.value,
        LeaseState.ACTIVE.value,
        LeaseState.CANCELLED.value,
    ]
    active_leases: Mapped[List["LeaseEntry"]] = relationship(
        "LeaseEntry",
        primaryjoin=f"and_(LeaseEntry.job_name==JobEntry.name, LeaseEntry.state.in_({active_states}))",
        order_by="LeaseEntry.id.desc()",
        overlaps="job,leases",
    )

    operations: Mapped[List["OperationEntry"]] = relationship("OperationEntry", back_populates="job")

    platform: Mapped[List["PlatformEntry"]] = relationship(
        "PlatformEntry", secondary=job_platform_association, back_populates="jobs"
    )

    __table_args__ = (
        Index(
            "ix_worker_start_timestamp",
            "worker_start_timestamp",
            unique=False,
            postgresql_where="worker_start_timestamp IS NOT null",
            sqlite_where="worker_start_timestamp IS NOT null",
        ),
        Index(
            "ix_worker_completed_timestamp",
            "worker_completed_timestamp",
            unique=False,
            postgresql_where="worker_completed_timestamp IS NOT null",
            sqlite_where="worker_completed_timestamp IS NOT null",
        ),
        Index(
            "ix_jobs_stage_property_label",
            "stage",
            "property_label",
            unique=False,
        ),
    )


class LeaseEntry(Base):
    __tablename__ = "leases"

    id: Mapped[int] = mapped_column(primary_key=True)
    job_name: Mapped[str] = mapped_column(ForeignKey("jobs.name", ondelete="CASCADE", onupdate="CASCADE"), index=True)
    status: Mapped[Optional[int]]
    state: Mapped[int]
    worker_name: Mapped[Optional[str]] = mapped_column(index=True, default=None)

    job: Mapped[JobEntry] = relationship(JobEntry, back_populates="leases")

    def to_protobuf(self) -> bots_pb2.Lease:
        lease = bots_pb2.Lease()
        lease.id = self.job_name

        if self.job.action is not None:
            action = Action()
            action.ParseFromString(self.job.action)
            lease.payload.Pack(action)
        else:
            lease.payload.Pack(string_to_digest(self.job.action_digest))

        lease.state = self.state  # type: ignore[assignment]
        if self.status is not None:
            lease.status.code = self.status
        return lease


class ClientIdentityEntry(Base):
    __tablename__ = "client_identities"
    __table_args__ = (UniqueConstraint("instance", "workflow", "actor", "subject"),)

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    instance: Mapped[str]
    workflow: Mapped[str]
    actor: Mapped[str]
    subject: Mapped[str]

    def __str__(self) -> str:
        return (
            f"ClientIdentity: [instance={self.instance} workflow={self.workflow}"
            f" actor={self.actor} subject={self.subject}]"
        )


class RequestMetadataEntry(Base):
    __tablename__ = "request_metadata"
    __table_args__ = (
        UniqueConstraint(
            "tool_name",
            "tool_version",
            "invocation_id",
            "correlated_invocations_id",
            "action_mnemonic",
            "target_id",
            "configuration_id",
            name="unique_metadata_constraint",
        ),
    )

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    tool_name: Mapped[Optional[str]]
    tool_version: Mapped[Optional[str]]
    invocation_id: Mapped[Optional[str]]
    correlated_invocations_id: Mapped[Optional[str]]
    action_mnemonic: Mapped[Optional[str]]
    target_id: Mapped[Optional[str]]
    configuration_id: Mapped[Optional[str]]


class OperationEntry(Base):
    __tablename__ = "operations"

    name: Mapped[str] = mapped_column(primary_key=True)
    cancelled: Mapped[bool] = mapped_column(default=False, nullable=False)
    tool_name: Mapped[Optional[str]]
    tool_version: Mapped[Optional[str]]
    invocation_id: Mapped[Optional[str]]
    correlated_invocations_id: Mapped[Optional[str]]

    job_name: Mapped[str] = mapped_column(ForeignKey("jobs.name", ondelete="CASCADE", onupdate="CASCADE"), index=True)
    job: Mapped[JobEntry] = relationship(JobEntry, back_populates="operations")

    client_identity_id: Mapped[Optional[int]] = mapped_column(ForeignKey("client_identities.id"))
    client_identity: Mapped[Optional[ClientIdentityEntry]] = relationship("ClientIdentityEntry")

    request_metadata_id: Mapped[Optional[int]] = mapped_column(ForeignKey(RequestMetadataEntry.id))
    request_metadata: Mapped[Optional[RequestMetadataEntry]] = relationship(RequestMetadataEntry)


class IndexEntry(Base):
    __tablename__ = "index"

    digest_hash: Mapped[str] = mapped_column(index=True, primary_key=True)
    digest_size_bytes: Mapped[bigint]
    accessed_timestamp: Mapped[datetime.datetime] = mapped_column(index=True)
    deleted: Mapped[bool] = mapped_column(server_default=false())
    inline_blob: Mapped[Optional[bytes]]


# This table is used to store the bot session state. It also stores the
# assigned leases, instead of making use of the 'leases' table through an
# SQLAlchemy relationship, as the 'leases' table is dependent on the type of
# data store selected, and might never be populated.
# TODO: We can now guarantee that `leases` exists, and should add a proper
# relationship in the data model instead of this.
class BotEntry(Base):
    __tablename__ = "bots"

    # Immutable data
    name: Mapped[str] = mapped_column(index=True, primary_key=True)
    bot_id: Mapped[str] = mapped_column(index=True)
    instance_name: Mapped[str]

    # Scheduling state
    bot_status: Mapped[int]
    lease_id: Mapped[Optional[str]]

    # Auditing data
    expiry_time: Mapped[datetime.datetime] = mapped_column(index=True)
    last_update_timestamp: Mapped[datetime.datetime] = mapped_column(index=True)

    job: Mapped[Optional[JobEntry]] = relationship(JobEntry, primaryjoin="foreign(BotEntry.lease_id) == JobEntry.name")


# This table is used by the SQLStorage CAS backend to store blobs
# in a database.
class BlobEntry(Base):
    __tablename__ = "blobs"

    digest_hash: Mapped[str] = mapped_column(primary_key=True)
    digest_size_bytes: Mapped[bigint]
    data: Mapped[bytes]


def digest_to_string(digest: Digest) -> str:
    return f"{digest.hash}/{digest.size_bytes}"


def string_to_digest(string: str) -> Digest:
    digest_hash, size_bytes = string.split("/", 1)
    return Digest(hash=digest_hash, size_bytes=int(size_bytes))
